# coding: utf-8
# ##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "17/05/2017"


import os
import logging
import shutil
from urllib.request import urlopen, ProxyHandler, build_opener

logger = logging.getLogger(__file__)
url_base = "http://www.edna-site.org/pub/xas/"


def DownloadDataset(dataset, output_folder, timeout, unpack=False):
    # create if needed path scan
    url = url_base + dataset

    logger.info("Trying to download scan %s, timeout set to %ss",
                dataset, timeout)
    dictProxies = {}
    if "http_proxy" in os.environ:
        dictProxies['http'] = os.environ["http_proxy"]
        dictProxies['https'] = os.environ["http_proxy"]
    if "https_proxy" in os.environ:
        dictProxies['https'] = os.environ["https_proxy"]
    if dictProxies:
        proxy_handler = ProxyHandler(dictProxies)
        opener = build_opener(proxy_handler).open
    else:
        opener = urlopen
    logger.info("wget %s" % url)
    data = opener(url, data=None, timeout=timeout).read()
    logger.info("Image %s successfully downloaded." % dataset)

    if not os.path.isdir(output_folder):
        os.mkdir(output_folder)

    try:
        archive_folder = os.path.join(output_folder, os.path.basename(dataset))
        with open(archive_folder, "wb") as outfile:
            outfile.write(data)
    except IOError:
        raise IOError("unable to write downloaded \
                        data to disk at %s" % archive_folder)

    if unpack is True:
        shutil.unpack_archive(archive_folder, extract_dir=output_folder, format='bztar')
        os.remove(archive_folder)




__authors__ = ["Bioinformatics Laboratory, University of Ljubljana", "H.Payno"]
__license__ = "[GNU GPL v3+]: https://www.gnu.org/licenses/gpl-3.0.en.html"
__date__ = "29/05/2017"


import sys
from xml.etree.ElementTree import parse
from collections import namedtuple
import ast
from ast import literal_eval
import json
import logging
import pickle
import base64
import importlib
from .scheme import Node, Scheme

logger = logging.getLogger(__name__)


"""This is a dictionnary to convert an OrangeWidget into a core process.
"""

"""Some widget are only used for visualization, and will be ignored"""

def scheme_load(_file, _type='ows', error_handler=None):
    assert _type in ("ows", "moml")
    if _type == 'ows':
        desc = parse_ows_stream(_file)
    elif _type == 'moml':
        desc = parse_moml_stream(_file)

    if error_handler is None:
        def error_handler(exc):
            raise exc

    nodes = []
    nodes_dict = {}

    for node_d in desc.nodes:
        callback = _get_callback(node_d)
        node = Node(id=node_d.id, callback=callback)
        nodes.append(node)
        nodes_dict[node.id] = node
        data = node_d.data
        if data:
            properties = loads(data.data, data.format)
            node.properties = properties
        else:
            node.properties = {}
        node.qualified_name = node_d.qualified_name

    for link_d in desc.links:
        upstream_node = nodes_dict[link_d.source_node_id]
        assert isinstance(upstream_node, Node)
        downstream_node = nodes_dict[link_d.sink_node_id]
        upstream_node.downstream_nodes.add(downstream_node)
        downstream_node.upstream_nodes.add(upstream_node)

    scheme = Scheme(nodes=nodes, links=desc.links)
    scheme.title = desc.title
    scheme.description = desc.description

    return scheme


def _get_callback(node):
    name = node.qualified_name
    sname = name.rsplit('.')
    assert (len(sname) > 1)
    class_name = sname[-1]
    del sname[-1]
    module_name = '.'.join(sname)
    m = importlib.import_module(module_name)

    class_ = getattr(m, class_name)
    if hasattr(class_, 'process_function'):
        return class_.process_function
    else:
        raise ValueError('No process function defined in the class core')


def loads(string, format):
    if format == "literal":
        return literal_eval(string)
    elif format == "json":
        return json.loads(string)
    elif format == "pickle":
        return pickle.loads(base64.decodebytes(string.encode('ascii')))
    else:
        raise ValueError("Unknown format")

# ---- TAKE back from Orange3 ---------

_scheme = namedtuple(
    "_scheme",
    ["title", "version", "description", "nodes", "links", "annotations"])

_node = namedtuple(
    "_node",
    ["id", "title", "name", "position", "project_name", "qualified_name",
     "version", "data"])

_data = namedtuple(
    "_data",
    ["format", "data"])

_link = namedtuple(
    "_link",
    ["id", "source_node_id", "sink_node_id", "source_channel", "sink_channel",
     "enabled"])

_annotation = namedtuple(
    "_annotation",
    ["id", "type", "params"])

_text_params = namedtuple(
    "_text_params",
    ["geometry", "text", "font"])

_arrow_params = namedtuple(
    "_arrow_params",
    ["geometry", "color"])

_nxNode = namedtuple(
    "_node",
    ["id", "class_", "properties", "data", "qualified_name"])

_nxNodeProperty = namedtuple(
    "_property",
    ["name", "class_", "value"])

_nxLink = namedtuple(
    "_nxLink",
    ["port", "relation"])

_nxRelation = namedtuple(
    "_relation",
    ["id", "class_", 'properties'])

def tuple_eval(source):
    """
    Evaluate a python tuple literal `source` where the elements are
    constrained to be int, float or string. Raise ValueError if not
    a tuple literal.

    >>> tuple_eval("(1, 2, "3")")
    (1, 2, '3')

    """
    if source is None:
        return None
    node = ast.parse(source, "<source>", mode="eval")

    if not isinstance(node.body, ast.Tuple):
        raise ValueError("%r is not a tuple literal" % source)

    if not all(isinstance(el, (ast.Str, ast.Num)) or
               # allow signed number literals in Python3 (i.e. -1|+1|-1.0)
               (isinstance(el, ast.UnaryOp) and
                isinstance(el.op, (ast.UAdd, ast.USub)) and
                isinstance(el.operand, ast.Num))
               for el in node.body.elts):
        raise ValueError("Can only contain numbers or strings")

    return literal_eval(source)


def parse_ows_etree_v_2_0(tree):
    scheme = tree.getroot()
    nodes, links, annotations = [], [], []

    # First collect all properties
    properties = {}
    for property in tree.findall("node_properties/properties"):
        node_id = property.get("node_id")
        format = property.get("format")
        if "data" in property.attrib:
            data = property.get("data")
        else:
            data = property.text
        properties[node_id] = _data(format, data)

    # Collect all nodes
    for node in tree.findall("nodes/node"):
        node_id = node.get("id")
        node = _node(
            id=node_id,
            title=node.get("title"),
            name=node.get("name"),
            position=tuple_eval(node.get("position", None)),
            project_name=node.get("project_name", None),
            qualified_name=node.get("qualified_name"),
            version=node.get("version", ""),
            data=properties.get(node_id, None)
        )
        nodes.append(node)

    for link in tree.findall("links/link"):
        params = _link(
            id=link.get("id"),
            source_node_id=link.get("source_node_id"),
            sink_node_id=link.get("sink_node_id"),
            source_channel=link.get("source_channel"),
            sink_channel=link.get("sink_channel"),
            enabled=link.get("enabled") == "true",
        )
        links.append(params)

    for annot in tree.findall("annotations/*"):
        if annot.tag == "text":
            rect = tuple_eval(annot.get("rect", "(0.0, 0.0, 20.0, 20.0)"))

            font_family = annot.get("font-family", "").strip()
            font_size = annot.get("font-size", "").strip()

            font = {}
            if font_family:
                font["family"] = font_family
            if font_size:
                font["size"] = int(font_size)

            annotation = _annotation(
                id=annot.get("id"),
                type="text",
                params=_text_params(rect, annot.text or "", font),
            )
        elif annot.tag == "arrow":
            start = tuple_eval(annot.get("start", "(0, 0)"))
            end = tuple_eval(annot.get("end", "(0, 0)"))
            color = annot.get("fill", "red")
            annotation = _annotation(
                id=annot.get("id"),
                type="arrow",
                params=_arrow_params((start, end), color)
            )
        annotations.append(annotation)
    return _scheme(
        version=scheme.get("version"),
        title=scheme.get("title", ""),
        description=scheme.get("description"),
        nodes=nodes,
        links=links,
        annotations=annotations
    )


def parse_moml_etree(tree):
    def convertToOrangeLinks(nx_links, nx_relations):
        def getLink(_id):
            if _id not in links:
                links[_id] = {"id": _id}
            return links[_id]

        links = {}
        # convert nx (.omlm) liks and relations to orange links
        for nx_link in nx_links:
            node_id, input_output = nx_link.port.split('.', -1)
            link = getLink(nx_link.relation)
            if input_output.lower() in ('output', 'other', 'true'):
                link["source_node_id"] = node_id
            else:
                link["sink_node_id"] = node_id

        for relation in nx_relations:
            assert relation.id in links
            links[relation.id]["source_channel"] = relation.class_
            links[relation.id]["sink_channel"] = relation.class_
            links[relation.id]["properties"] = relation.properties

        orangeLinks = []
        for linkid, link in links.items():
            l = _link(id=link['id'],
                      source_node_id=link["source_node_id"],
                      sink_node_id=link["sink_node_id"],
                      source_channel=link["source_channel"],
                      sink_channel=link["sink_channel"],
                      enabled=True)
            orangeLinks.append(l)

        return orangeLinks

    nodes, nx_links, nx_relations = [], [], []

    # Collect all nodes
    for node in tree.findall(".//entity"):
        node_id = node.get('name')
        node_class = node.get('class')
        node_properties = []
        for property in node.findall("property"):
            _property = _nxNodeProperty(name=property.get("name"),
                                        class_=property.get("class"),
                                        value=property.get("value"))
            node_properties.append(_property)
        nodes.append(_nxNode(id=node_id,
                             class_=node_class,
                             properties=node_properties,
                             data=None,
                             qualified_name=node_class))

    # collect all nx links
    for link in tree.findall("link"):
        _my_link = _nxLink(port=link.get("port"),
                           relation=link.get("relation"))
        nx_links.append(_my_link)

    # collect all nx relations
    for relation in tree.findall("relation"):
        relation_properties = []
        for property in relation.findall("property"):
            _property = _nxNodeProperty(name=property.get("name"),
                                        class_=property.get("class"),
                                        value=property.get("value"))
            relation_properties.append(_property)

        _relation = _nxRelation(id=relation.get("name"),
                                class_=relation.get("class"),
                                properties=relation_properties)
        nx_relations.append(_relation)

    links = convertToOrangeLinks(nx_links=nx_links, nx_relations=nx_relations)

    return _scheme(
        version="nx 0.1",
        title=("undefined"),
        description=None,
        nodes=nodes,
        links=links,
        annotations=None
    )


def parse_ows_stream(stream):
    doc = parse(stream)
    scheme_el = doc.getroot()
    version = scheme_el.get("version", None)
    if version is None:
        # Fallback: check for "widgets" tag.
        if scheme_el.find("widgets") is not None:
            version = "1.0"
        else:
            log.warning("<scheme> tag does not have a 'version' attribute")
            version = "2.0"

    if version == "1.0":
        return parse_ows_etree_v_1_0(doc)
    elif version == "2.0":
        return parse_ows_etree_v_2_0(doc)
    else:
        raise ValueError('unrecognize scheme definition version')


def parse_moml_stream(stream):
    doc = parse(stream)
    scheme = parse_moml_etree(doc)
    return scheme


def resolve_replaced(scheme_desc, registry):
    widgets = registry.widgets()
    nodes_by_id = {}  # type: Dict[str, _node]
    replacements = {}
    replacements_channels = {}  # type: Dict[str, Tuple[dict, dict]]
    # collect all the replacement mappings
    for desc in widgets:  # type: WidgetDescription
        if desc.replaces:
            for repl_qname in desc.replaces:
                replacements[repl_qname] = desc.qualified_name

        input_repl = {}
        for idesc in desc.inputs or []:  # type: InputSignal
            for repl_qname in idesc.replaces or []:  # type: str
                input_repl[repl_qname] = idesc.name
        output_repl = {}
        for odesc in desc.outputs:  # type: OutputSignal
            for repl_qname in odesc.replaces or []:  # type: str
                output_repl[repl_qname] = odesc.name
        replacements_channels[desc.qualified_name] = (input_repl, output_repl)

    # replace the nodes
    nodes = scheme_desc.nodes
    for i, node in list(enumerate(nodes)):
        if not registry.has_widget(node.qualified_name) and \
                node.qualified_name in replacements:
            qname = replacements[node.qualified_name]
            desc = registry.widget(qname)
            nodes[i] = node._replace(qualified_name=desc.qualified_name,
                                     project_name=desc.project_name)
        nodes_by_id[node.id] = nodes[i]

    # replace links
    links = scheme_desc.links
    for i, link in list(enumerate(links)):  # type: _link
        nsource = nodes_by_id[link.source_node_id]
        nsink = nodes_by_id[link.sink_node_id]

        _, source_rep = replacements_channels.get(
            nsource.qualified_name, ({}, {}))
        sink_rep, _ = replacements_channels.get(
            nsink.qualified_name, ({}, {}))

        if link.source_channel in source_rep:
            link = link._replace(
                source_channel=source_rep[link.source_channel])
        if link.sink_channel in sink_rep:
            link = link._replace(
                sink_channel=sink_rep[link.sink_channel])
        links[i] = link

    return scheme_desc._replace(nodes=nodes, links=links)


def main(ows_file):
    scheme = _GUIFreeScheme()
    with open(ows_file, "rb") as f:
        scheme_load(scheme, f)

    scheme.node_by_id['0'].start()


if __name__ == "__main__":
    if len(sys.argv) != 2 or not sys.argv[1].lower().endswith('.ows'):
        raise ValueError('need a .ows file to process the scheme')

    main(sys.argv[1])



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/07/2019"


import functools
import logging

from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
import Orange.data
from PyMca5.PyMcaGui.physics.xas.XASNormalizationParameters import \
    XASNormalizationParameters
from silx.gui import qt
from silx.gui.plot import LegendSelector
from est.gui.XasObjectViewer import (_plot_edge, _plot_norm, _plot_post_edge,
                                     _plot_pre_edge)
import est.core.process.pymca.normalization
from orangecontrib.est.process import _ProcessForOrangeMixIn
from orangecontrib.est.process import ProcessRunnable
from est.core.types import XASObject
from est.gui.XasObjectViewer import XasObjectViewer, ViewType
from orangecontrib.est.progress import QProgress
from orangecontrib.est.utils import Converter
from ..container import _ParameterWindowContainer

_logger = logging.getLogger(__file__)


class NormalizationWindow(qt.QMainWindow):
    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # xas object viewer
        mapKeys = ['mu', 'NormalizedMu', 'NormalizedSignal', 'NormalizedBackground']
        self.xasObjViewer = XasObjectViewer(mapKeys=mapKeys)
        self.xasObjViewer._spectrumViews[0]._plot.getXAxis().setLabel("Energy (eV)")
        self.xasObjViewer._spectrumViews[0]._plot.getYAxis().setLabel("Absorption (a.u.)")
        self.setCentralWidget(self.xasObjViewer)
        self._pymcaWindow = _ParameterWindowContainer(parent=self,
                                                      parametersWindow=XASNormalizationParameters)
        dockWidget = qt.QDockWidget(parent=self)

        # pymca window
        dockWidget.setWidget(self._pymcaWindow)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        # legend selector
        self.legendDockWidget = LegendSelector.LegendsDockWidget(parent=self,
                                                                 plot=self.xasObjViewer._spectrumViews[0]._plot)
        self.legendDockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        self.legendDockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.legendDockWidget)

        # volume key selection
        self.addDockWidget(qt.Qt.RightDockWidgetArea,
                           self.xasObjViewer._mapView.keySelectionDocker)

        # plot settings
        for ope in (_plot_edge, _plot_norm, _plot_post_edge, _plot_pre_edge):
            self.xasObjViewer._spectrumViews[0].addCurveOperation(ope)

        self.setWindowFlags(qt.Qt.Widget)

        # connect signal / slot
        self.xasObjViewer.viewTypeChanged.connect(self._updateLegendView)

        # set up
        self._updateLegendView()

    def getNCurves(self):
        return len(self.xasObjViewer._spectrumViews._plot.getAllCurves())

    def _updateLegendView(self):
        index, viewType = self.xasObjViewer.getViewType()
        self.legendDockWidget.setVisible(viewType is ViewType.spectrum)
        self.xasObjViewer._mapView.keySelectionDocker.setVisible(viewType is ViewType.map)


class NormalizationOW(_ProcessForOrangeMixIn, OWWidget):
    """
    Widget used for signal extraction
    """
    name = "normalization"
    id = "orange.widgets.xas.pymca.normalization"
    description = "Progress spectra normalization"
    icon = "icons/normalization.png"
    priority = 1
    category = "esrfWidgets"
    keywords = ["spectroscopy", "normalization"]

    want_main_area = True
    resizing_enabled = True

    process_function = est.core.process.pymca.normalization.PyMca_normalization

    _pymcaSettings = Setting(dict())
    """Store the configuration of the PyMca XASClass"""

    class Inputs:
        xas_obj = Input('xas_obj', XASObject, default=True)
        # simple compatibility for some Orange widget and especialy the
        # 'spectroscopy add-on'
        data_table = Input('Data', Orange.data.Table)

    class Outputs:
        res_xas_obj = Output('xas_obj', XASObject)
        # by default we want to avoid sending 'Orange.data.Table' to avoid
        # loosing the XASObject flow process and results.

    def __init__(self):
        super().__init__()
        self._latest_xas_obj = None
        self._window = NormalizationWindow(parent=self)
        layout = gui.vBox(self.mainArea, 'normalization').layout()
        layout.addWidget(self._window)
        self._window.xasObjViewer.setWindowTitle('spectra')

        # manage settings
        if self._pymcaSettings != dict():
            self._window._pymcaWindow.setParameters(self._pymcaSettings)

        # connect signals / slots
        _sig = self._window._pymcaWindow.sigChanged.connect(self._updateProcess)

    def _updateProcess(self):
        self._update_settings()
        if self._latest_xas_obj:
            self.process(self._latest_xas_obj)

    def _update_settings(self):
        self._pymcaSettings = self._window._pymcaWindow.getParameters()

    @Inputs.data_table
    def processFrmDataTable(self, data_table):
        if data_table is None:
            return
        self.process(Converter.toXASObject(data_table=data_table))

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return

        if not self._canProcess():
            _logger.warning('There is some processing on going already, will'
                            'not process the new dataset')

        self._latest_xas_obj = xas_obj.copy(create_h5_file=True)
        self._startProcess()

        # setup the normalization process
        process_obj = QPyMca_normalization()
        process_obj._advancement.sigProgress.connect(self._setProgressValue)
        process_obj.setProperties({'_pymcaSettings': self._window._pymcaWindow.getParameters()})

        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(process_obj=process_obj, xas_obj=self._latest_xas_obj)
        self._callback_finish = functools.partial(self._endProcess,
                                                  self._latest_xas_obj)
        thread.finished.connect(self._callback_finish)
        # start processing
        thread.start(priority=qt.QThread.LowPriority)


class QPyMca_normalization(
        est.core.process.pymca.normalization.PyMca_normalization):
    """
    Normalization able to give advancement using qt.Signal and QThreadPool
    """

    def __init__(self):
        est.core.process.pymca.normalization.PyMca_normalization.__init__(self)
        self._advancement = QProgress('normalization')

    def _pool_process(self, xas_obj):
        self.pool = qt.QThreadPool()
        self.pool.setMaxThreadCount(5)
        for spectrum in xas_obj.spectra:
            runnable = ProcessRunnable(fct=est.core.process.pymca.normalization.process_spectr_norm,
                                       spectrum=spectrum,
                                       configuration=xas_obj.configuration,
                                       callback=self._advancement.increaseAdvancement)
            self.pool.start(runnable)
        self.pool.waitForDone()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/12/2019"

import logging
from datetime import datetime

import h5py
import numpy
from silx.io import utils
from silx.io.dictdump import dicttoh5
from silx.io.url import DataUrl
from silx.utils.enum import Enum

try:
    from est.io.utils.pymca import read_spectrum as pymca_read_spectrum
except ImportError:
    has_pymca = False
else:
    has_pymca = True
try:
    from est.io.utils.larch import read_ascii as larch_read_ascii
except ImportError:
    has_larch = False
else:
    has_larch = True

_logger = logging.getLogger(__name__)


class InputType(Enum):
    dat_spectrum = '*.dat'    # contains one spectrum
    hdf5_spectra = '*.h5'     # contains several spectrum
    xmu_spectrum = '*.xmu'  # contains one spectrum


def read_xas(spectra_url, channel_url, config_url=None):
    """
    Read the given spectra url and the config url if any
    
    :param Union[DataUrl, str] spectra_url: 
    :param DataUrl config_url: 
    :return: spectra, energy, configuration
    """
    def get_url(original_url, name):
        url_ = original_url
        if type(url_) is str:
            try:
                url_ = DataUrl(path=url_)
            except:
                url_ = DataUrl(file_path=url_, scheme='PyMca')

        if not isinstance(url_, DataUrl):
            raise TypeError('given input for, ', name, 'is invalid')
        return url_

    _spectra_url = get_url(original_url=spectra_url, name='spectra')
    _energy_url = get_url(original_url=channel_url, name='energy')
    _config_url = config_url
    if type(_config_url) is str and _config_url == '':
        _config_url = None
    if not (_config_url is None or isinstance(_config_url, DataUrl)):
        raise TypeError('given input for configuration is invalid')

    def load_data(data_url, name):
        if data_url is None:
            return None
        if data_url.scheme() in ('PyMca', 'PyMca5'):
            if has_pymca is False:
                _logger.warning('Requires PyMca to load data from '
                                '%s' % data_url.path())
                return None
            else:
                assert name in ('spectra', 'energy')
                energy, mu = pymca_read_spectrum(data_url.file_path())
                if name == 'spectra':
                    return mu.reshape(mu.shape[0], 1, 1)
                else:
                    return energy
        elif data_url.scheme() in ('larch', 'xraylarch'):
            if has_larch is False:
                _logger.warning('Requires larch to load data from '
                                '%s' % data_url.path())
                return None
            else:
                assert name in ('spectra', 'energy')
                energy, mu = larch_read_ascii(xmu_file=data_url.file_path())
                if name == 'spectra':
                    return mu.reshape(mu.shape[0], 1, 1)
                else:
                    return energy
        else:
            if data_url.is_valid():
                try:
                    return utils.get_data(data_url)
                except ValueError as e:
                    _logger.error(e)
            else:
                _logger.warning('invalid url for', name, ',  will not load it')

    spectra = load_data(_spectra_url, name='spectra')
    energy = load_data(_energy_url, name='energy')
    configuration = load_data(_config_url, name='configuration')

    return (spectra, energy, configuration)


def write_xas_proc(h5_file, entry, process, data, processing_order,
                   data_path='/', overwrite=True):
    """
    Write a xas :class:`.Process` into .h5

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param process: process executed
    :type: :class:`.Process`
    :param data: process result data
    :type: numpy.ndarray
    :param processing_order: processing order of treatment
    :type: int
    :param data_path: path to store the data
    :type: str
    """
    process_name = 'xas_process_' + str(processing_order)
    # write the xasproc
    with h5py.File(h5_file) as h5f:
        nx_entry = h5f.require_group('/'.join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"

        nx_process = nx_entry.require_group(process_name)
        nx_process.attrs['NX_class'] = "NXprocess"
        if overwrite:
            for key in ('program', 'version', 'date', 'processing_order',
                        'class_instance'):
                if key in nx_process:
                    del nx_process[key]
        nx_process['program'] = process.program_name()
        nx_process['version'] = process.program_version()
        nx_process['date'] = datetime.now().replace(microsecond=0).isoformat()
        nx_process['processing_order'] = numpy.int32(processing_order)
        _class = process.__class__
        nx_process['class_instance'] = '.'.join((_class.__module__,
                                                 _class.__name__))

        nx_data = nx_entry.require_group('data')
        nx_data.attrs['NX_class'] = "NXdata"
        nx_data.attrs['signal'] = 'data'

        def get_interpretation(mydata):
            if isinstance(mydata, numpy.ndarray):
                if mydata.ndim is 1:
                    return 'spectrum'
                elif mydata.ndim in (2, 3):
                    return 'image'
            return None

        if isinstance(data, numpy.ndarray):
            data_ = {'data': data}
        else:
            data_ = data
        for key, value in data_.items():
            nx_process[key] = value
            interpretation = get_interpretation(value)
            if interpretation:
                nx_process[key].attrs['interpretation'] = interpretation
        nx_process_path = str(nx_process.name)

    if process.getConfiguration() is not None:
        dicttoh5(process.getConfiguration(),
                 h5file=h5_file,
                 h5path='/'.join((nx_process_path, 'configuration')),
                 overwrite_data=True,
                 mode='a')


def write_xas(h5_file, entry, energy, mu, sample=None, start_time=None,
              data_path='/', title=None, definition=None, overwrite=True):
    """
    Write raw date in nexus format

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param sample: definition of the sample
    :type: :class:`.Sample`
    :param energy: beam energy (1D)
    :type: numpy.ndarray
    :param mu: beam absorption (2D)
    :type: numpy.ndarray
    :param start_time:
    :param str data_path:
    :param str title: experiment title
    :param str definition: experiment definition
    """
    with h5py.File(h5_file) as h5f:
        nx_entry = h5f.require_group('/'.join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"

        # store energy
        nx_monochromator = nx_entry.require_group('monochromator')
        nx_monochromator.attrs['NX_class'] = "NXmonochromator"
        if overwrite and 'energy' in nx_monochromator:
            del nx_monochromator['energy']
        nx_monochromator['energy'] = energy
        nx_monochromator['energy'].attrs['interpretation'] = 'spectrum'
        nx_monochromator['energy'].attrs['NX_class'] = "NXdata"

        # store absorbed beam
        nx_absorbed_beam = nx_entry.require_group('absorbed_beam')
        nx_absorbed_beam.attrs['NX_class'] = "NXdetector"
        if overwrite and 'data' in nx_absorbed_beam:
            del nx_absorbed_beam['data']
        nx_absorbed_beam['data'] = mu
        nx_absorbed_beam['data'].attrs['interpretation'] = 'image'
        nx_absorbed_beam['data'].attrs['NX_class'] = "NXdata"

        if sample:
            nx_sample = nx_entry.require_group('sample')
            nx_sample.attrs['NX_class'] = "NXsample"
            if overwrite and 'name' in nx_sample:
                del nx_sample['name']
            nx_sample['name'] = sample.name

        nx_data = nx_entry.require_group('data')
        nx_data.attrs['NX_class'] = "NXdata"
        # create some link on data
        if overwrite and 'energy' in nx_data:
            del nx_data['energy']
        nx_data['energy'] = h5py.SoftLink(nx_monochromator['energy'].name)
        if overwrite and 'absorbed_beam' in nx_data:
            del nx_data['absorbed_beam']
        nx_data['absorbed_beam'] = h5py.SoftLink(nx_absorbed_beam['data'].name)

        if start_time is not None:
            if overwrite and 'start_time' in nx_entry:
                del nx_entry['start_time']
            nx_entry['start_time'] = start_time
        if title is not None:
            if overwrite and 'title' in nx_entry:
                del nx_entry['title']
            nx_entry['title'] = title
        if definition is not None:
            if overwrite and 'definition' in nx_entry:
                del nx_entry['definition']
            nx_entry['definition'] = definition


def get_xasproc(h5_file, entry):
    """
    Return the list of all NXxasproc existing at the data_path level
    
    :param str h5_file: hdf5 file
    :param str entry: data location
     
    :return: 
    :rtype: list
    """
    def copy_nx_xas_process(h5_group):
        """copy base information from nx_xas_process"""
        res = {}
        res['_h5py_path'] = h5_group.name
        relevant_keys = ('program', 'version', 'data', 'parameters',
                         'processing_order', 'configuration', 'class_instance')
        from silx.io.dictdump import h5todict
        for key in h5_group.keys():
            # for now we don't want to copy the numpy array (data)
            if key in relevant_keys:
                if key == 'configuration':
                    config_path = '/'.join((h5_group.name, 'configuration'))
                    res[key] = h5todict(h5_file, config_path)
                else:
                    res[key] = h5_group[key][...]
        return res

    res = []
    with h5py.File(h5_file) as h5f:
        try:
            root_group = h5f[entry]
        except KeyError:
            _logger.warning(entry + ' does not exist in ' + h5_file)
        else:
            for key in root_group.keys():
                elmt = root_group[key]
                if hasattr(elmt, 'attrs') and 'NX_class' in elmt.attrs:
                    if elmt.attrs['NX_class'] == 'NXprocess':
                        nx_xas_proc = copy_nx_xas_process(elmt)
                        if len(nx_xas_proc) == 0:
                            _logger.warning('one xas process was not readable '
                                            'from the hdf5 file at:' + key)
                        else:
                            res.append(nx_xas_proc)
    return res


if __name__ == '__main__':
    import os
    from est.core.process.pymca.normalization import PyMca_normalization
    from est.core.process.pymca.exafs import PyMca_exafs
    from est.core.types import Sample
    h5_file = 'test_xas_123.h5'
    if os.path.exists(h5_file):
        os.remove(h5_file)
    sample = Sample(name='mysample')
    data = numpy.random.rand(256*20*10)
    data = data.reshape((256, 20, 10))
    process_data = numpy.random.rand(256*20*10).reshape((256, 20, 10))
    energy = numpy.linspace(start=3.25, stop=3.69, num=256)

    write_xas(h5_file=h5_file, entry='scan1', sample=sample, energy=energy,
              mu=data, )

    process_norm = PyMca_normalization()
    write_xas_proc(h5_file=h5_file, entry='scan1', process=process_norm,
                   data=process_data, processing_order=1)
    process_exafs = PyMca_exafs()
    process_data2 = numpy.random.rand(256*20*10).reshape((256, 20, 10))
    write_xas_proc(h5_file=h5_file, entry='scan1', process=process_exafs,
                   data=process_data2, processing_order=2)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""module for process base class"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/08/2019"


from .progress import Progress
from est.core.types import XASObject
import logging
_logger = logging.getLogger(__name__)


_DEBUG = True
if _DEBUG:
    from est.core.types import Spectrum
    import numpy


class Process(object):
    def __init__(self, name):
        assert type(name) is str
        self._name = name
        self._advancement = Progress(name=name)
        self.__stop = False
        """flag to notice when a end of process is required"""
        self._settings = {}
        # configuration
        self._callbacks = []

    @property
    def name(self):
        return self._name

    def stop(self):
        self.__stop = True

    @property
    def advancement(self):
        return self._advancement

    @advancement.setter
    def advancement(self, advancement):
        assert isinstance(advancement, Progress)
        self._advancement = advancement

    @property
    def callbacks(self):
        return self._callbacks

    @staticmethod
    def getXasObject(xas_obj):
        if isinstance(xas_obj, dict):
            _xas_obj = XASObject.from_dict(xas_obj)
        else:
            _xas_obj = xas_obj
        assert isinstance(_xas_obj, XASObject)
        if _xas_obj.n_spectrum > 0:
            for spectrum in _xas_obj.spectra:
                assert isinstance(spectrum, Spectrum)
            assert isinstance(spectrum.energy, numpy.ndarray)
            assert isinstance(spectrum.mu, numpy.ndarray)
        assert isinstance(_xas_obj, XASObject)
        return _xas_obj

    def program_name(self):
        """Name of the program used for this processing"""
        raise NotImplementedError('Base class')

    def program_version(self):
        """version of the program used for this processing"""
        raise NotImplementedError('Base class')

    def definition(self):
        """definition of the process"""
        raise NotImplementedError('Base class')

    def getConfiguration(self):
        """

        :return: configuration of the process
        :rtype: dict
        """
        if len(self._settings) > 0:
            return self._settings
        else:
            return None

    def setConfiguration(self, configuration):
        self._settings = configuration

    def register_process(self, xas_obj, data_keys):
        """
        Store the current process in the linked h5 file if any,
        output data stored will be the one defined by the data_keys

        :param xas_obj: object for which we want to save the treatment
        :type: :class:`.XASObject`
        :param tuple data_keys: keys of the id to save
        :param str relative_to: x value
        """
        if xas_obj.has_linked_file():
            _data = {}
            for key in data_keys:
                relative_to = None
                if key in ('flat', 'fpp', 'f2', 'dmude', 'norm', 'norm_area',
                           'post_edge', 'bkg', 'energy', 'mback_mu',
                           'norm_mback', 'Energy', 'Mu', "NormalizedEnergy",
                           "NormalizedMu", "NormalizedSignal", "EXAFSKValues",
                           "EXAFSSignal"):
                    relative_to = 'energy'
                    use = '_spectra_volume'
                elif key in ('chir_re', 'chir_im', 'chir_mag', 'r'):
                    relative_to = 'r'
                    use = '_spectra_volume'
                elif key in ('ft.radius', 'ft.intensity', 'ft.imaginary'):
                    relative_to = 'radius'
                    use = '_list_res_ft'
                if use == '_spectra_volume':
                    # if we can display the result as a numpy.array 3d
                    try:
                        _data[key] = XASObject._spectra_volume(xas_obj.spectra,
                                                               key=key,
                                                               dim_1=xas_obj.dim1,
                                                               dim_2=xas_obj.dim2,
                                                               relative_to=relative_to)
                    except KeyError:
                        mess = '%s: unable to store %s, parameter not found' % (self.name, key)
                        _logger.error(mess)

                elif use == '_list_res_ft':
                    res = []
                    key_ = key.split('.')[-1]
                    for spectrum in xas_obj.spectra:
                        res.append(getattr(spectrum.ft, key_))
                    _data[key] = res
                else:
                    raise ValueError()
            xas_obj.register_processing(process=self, data=_data)

    def addCallback(self, callback):
        self._callbacks.append(callback)


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


from silx.io.url import DataUrl
import est.io
import copy
import numpy
import logging
import h5py
import tempfile
import os
import shutil
try:
    import larch
except ImportError:
    _Spectrum_Base = object
    _has_larch = False
else:
    from larch.symboltable import Group
    _Spectrum_Base = Group
    _has_larch = True


_logger = logging.getLogger(__name__)


class XASObject(object):
    """Base class of XAS

    :param spectra: absorbed beam as a list of :class:`.Spectrum` or a
                    numpy.ndarray
    :type: Union[numpy.ndarray, list]
    :param energy: beam energy
    :type: numpy.ndarray of one dimension
    :param dict configuration: configuration of the different process
    :param int dim1: first dimension of the spectra
    :param int dim2: second dimension of the spectra
    :param str name: name of the object. Will be used for the hdf5 entry
    :param bool keep_process_flow: if True then will keep the trace of the set
                                   of process applied to the XASObject into a
                                   hdf5 file.
    """

    def __init__(self, spectra=None, energy=None, configuration=None, dim1=None,
                 dim2=None, name='scan1', keep_process_flow=True):
        self.__channels = None
        self.__spectra = []
        self.__energy = None
        self.__dim1 = 0
        self.__dim2 = 0
        self.__processing_index = 0
        self.__h5_file = None
        self.__entry_name = name
        self.spectra = (energy, spectra, dim1, dim2)
        self.configuration = configuration
        if keep_process_flow is True:
            self.__h5_file = os.path.join(tempfile.mkdtemp(), '_'.join((name, '_flow.h5')))
            self.link_to_h5(self.__h5_file)

    @property
    def entry(self):
        return self.__entry_name

    @property
    def spectra(self):
        return self.__spectra

    @spectra.setter
    def spectra(self, energy_spectra):
        energy, spectra, dim1, dim2 = energy_spectra
        if spectra is None:
            self.__spectra = []
            self.__energy = energy
        else:
            assert energy is not None
            self.__spectra.clear()
            assert isinstance(spectra, (list, tuple, numpy.ndarray))
            if isinstance(spectra, numpy.ndarray):
                assert spectra.ndim is 3
                self.__dim1 = spectra.shape[1]
                self.__dim2 = spectra.shape[2]
                for y_i_spectrum in range(spectra.shape[1]):
                    for x_i_spectrum in range(spectra.shape[2]):
                        self.addSpectrum(Spectrum(energy=energy,
                                                  mu=spectra[:, y_i_spectrum, x_i_spectrum]))
            else:
                if dim1 is None or dim2 is None:
                    raise ValueError(
                        'If you want to set spectra from a list/tuple '
                        'of Spectrum you should specify the spectra '
                        'dimensions')
                self.__dim1 = dim1
                self.__dim2 = dim2
                for spectrum in spectra:
                    assert isinstance(spectrum, Spectrum)
                    self.addSpectrum(spectrum)
        self.energy = energy

    def _setSpectra(self, spectra):
        self.__spectra = spectra

    def getSpectrum(self, dim1_idx, dim2_idx):
        """Util function to access the spectrum at dim1_idx, dim2_idx"""
        assert dim1_idx < self.dim1
        assert dim2_idx < self.dim2
        global_idx = dim1_idx * self.dim2 + dim2_idx
        assert global_idx < len(self.spectra)
        assert global_idx >= 0
        return self.spectra[global_idx]

    def addSpectrum(self, spectrum):
        self.__spectra.append(spectrum)

    @property
    def dim1(self):
        return self.__dim1

    def forceDim1(self, value):
        assert type(value) is int
        self.__dim1 = value

    def forceDim2(self, value):
        assert type(value) is int
        self.__dim2 = value

    @property
    def dim2(self):
        return self.__dim2

    @property
    def energy(self):
        return self.__energy

    @energy.setter
    def energy(self, energy):
        self.__energy = energy
        if len(self.__spectra) > 0:
            if len(self.__spectra[0].energy) != len(energy):
                _logger.warning('spectra and energy have incoherent dimension')

    @property
    def configuration(self):
        return self.__configuration

    @configuration.setter
    def configuration(self, configuration):
        assert configuration is None or isinstance(configuration, dict)
        self.__configuration = configuration or {}

    def to_dict(self, with_process_details=True):
        """convert the XAS object to a dict

        By default made to simply import raw data.

        :param with_process_details: used to embed a list of spectrum with
                                     intermediary result instead of only raw mu.
                                     This is needed especially for the
                                     pushworkflow actors to keep a trace of the
                                     processes.
        :type: bool
        """
        def get_list_spectra():
            res = []
            for spectrum in self.spectra:
                res.append(spectrum.to_dict())
            return res
        res = {
            'configuration': self.configuration,
            'spectra': XASObject._spectra_volume(spectra=self.spectra,
                                                 key='mu',
                                                 dim_1=self.dim1,
                                                 dim_2=self.dim2),
            'energy': self.energy,
            'dim1': self.dim1,
            'dim2': self.dim2,
        }
        if with_process_details is True:
            res['spectra'] = get_list_spectra()
            res['linked_h5_file'] = self.linked_h5_file
            res['current_processing_index'] = self.__processing_index

        return res

    def _spectra_to_dict(self):
        spectra_dict = {}
        for i_spectrum, spectrum in enumerate(self.spectra):
            assert isinstance(spectrum, Spectrum)
            spectra_dict[str(i_spectrum) + '_spectrum'] = spectrum.to_dict()
        return spectra_dict

    def absorbed_beam(self):
        return XASObject._spectra_volume(spectra=self.spectra,
                                         key='mu',
                                         dim_1=self.dim1,
                                         dim_2=self.dim2)

    @staticmethod
    def _spectra_volume(spectra, key, dim_1, dim_2, relative_to='energy'):
        """Convert a list of spectra (mu) to a numpy array.
        ..note: only convert raw data for now"""
        def get_param(object, name):
            if hasattr(object, name):
                return getattr(object, name)
            else:
                return object[name]

        def get_relative_to_value(relative):
            paths_ = relative.split('.')
            value = getattr(spectra[0], paths_[0])
            for path_ in paths_[1:]:
                value = getattr(value, path_)
            return value

        if len(spectra) is 0:
            return None
        else:
            assert len(spectra) == dim_1 * dim_2
            if relative_to:
                assert get_relative_to_value(relative_to) is not None
                array = numpy.zeros((len(get_relative_to_value(relative_to)),
                                     dim_1 * dim_2))
            else:
                array = numpy.zeros(dim_1 * dim_2)
            for i_spectrum, spectrum in enumerate(spectra):
                try:
                    if '.' in key:
                        subkeys = key.split('.')
                        key_ = subkeys[-1]
                        subkeys = subkeys[:-1]
                        value = get_param(spectrum, subkeys[0])
                        for subkey in subkeys[1:]:
                            value = get_param(value, subkey)
                        value = get_param(value, key_)
                    else:
                        value = get_param(spectrum, key)
                except:
                    _logger.error('fail to access to', key)
                else:
                    if _has_larch and isinstance(value, larch.symboltable.Group):
                        _logger.info('pass larch details, not managed for now')
                        continue
                    if relative_to is not None:
                        array[:, i_spectrum] = value
                    else:
                        array[i_spectrum] = value
            if relative_to is not None:
                return array.reshape((len(get_relative_to_value(relative_to)),
                                      dim_1, dim_2))
            else:
                return array.reshape(dim_1, dim_2)

    def load_frm_dict(self, ddict):
        """load XAS values from a dict"""
        contains_config_spectrum = 'configuration' in ddict or 'spectra' in ddict
        """The dict can be on the scheme of the to_dict function, containing
        the spectra and the configuration. Otherwise we consider it is simply
        the spectra"""
        if 'configuration' in ddict:
            self.configuration = ddict['configuration']
        if 'spectra' in ddict:
            spectra = ddict['spectra']
            if not isinstance(spectra, numpy.ndarray):
                new_spectra = []
                for spectrum in spectra:
                    assert isinstance(spectrum, dict)
                    new_spectra.append(Spectrum.from_dict(spectrum))
                spectra = new_spectra
        else:
            spectra = None
        if 'energy' in ddict:
            energy = ddict['energy']
        else:
            energy = None
        if 'dim1' in ddict:
            dim1 = ddict['dim1']
        else:
            dim1 = None
        if 'dim2' in ddict:
            dim2 = ddict['dim2']
        else:
            dim2 = None
        if 'linked_h5_file' in ddict:
            assert 'current_processing_index' in ddict
            self.link_to_h5(ddict['linked_h5_file'])
            self.__processing_index = ddict['current_processing_index']

        self.spectra = (energy, spectra, dim1, dim2)

        if not contains_config_spectrum:
            self.spectrum = ddict
        return self

    @staticmethod
    def from_dict(ddict):
        return XASObject().load_frm_dict(ddict=ddict)

    @staticmethod
    def from_file(h5_file, entry='scan1', spectra_path='data/absorbed_beam',
                  energy_path='data/energy', configuration_path='configuration'):
        # load only mu and energy from the file
        spectra_url = DataUrl(file_path=h5_file,
                              data_path='/'.join((entry, spectra_path)),
                              scheme='silx')
        energy_url = DataUrl(file_path=h5_file,
                             data_path='/'.join((entry, energy_path)),
                             scheme='silx')
        if configuration_path is None:
            config_url = None
        else:
            config_url = DataUrl(file_path=h5_file,
                                 data_path='/'.join((entry, configuration_path)),
                                 scheme='silx')
        sp, en, conf = xas.io.read_xas(spectra_url=spectra_url,
                                       channel_url=energy_url,
                                       config_url=config_url)
        return XASObject(spectra=sp, energy=en, configuration=conf)

    def dump(self, h5_file):
        """dump the XAS object to a file_path within the Nexus format"""
        from est.core.io import estWriter
        writer = XASWriter()
        # dump raw data
        writer.output_file = h5_file
        writer.dump_xas(xas_obj=self, write_process=True)

    def copy(self, create_h5_file=False):
        """

        :param bool create_h5_file: True if we want to duplicate the h5 file
                                 This is the case for orange widget for example
                                 where some reprocessing can append and each
                                 process need to keep a clear history of
                                 processes, whith no knowledge of next
                                 processing.
        """
        # To have dedicated h5 file we have to create one new h5 file for
        # for each process. For now there is no way to do it differently
        res = copy.copy(self)
        if create_h5_file is True:
            new_h5_file = os.path.join(tempfile.mkdtemp(), '_'.join((self.__entry_name, '_flow.h5')))
            if self.has_linked_file() and os.path.exists(self.linked_h5_file):
                shutil.copy(src=self.linked_h5_file, dst=new_h5_file)
            res.link_to_h5(new_h5_file)
        return res


    def __eq__(self, other):
        return (isinstance(other, XASObject) and
                numpy.array_equal(self.energy, other.energy) and
                self.dim1 == other.dim1 and
                self.dim2 == other.dim2 and
                self.configuration == other.configuration and
                self.spectra_equal(self.spectra, other.spectra))

    @staticmethod
    def spectra_equal(spectra1, spectra2):
        if len(spectra1) != len(spectra2):
            return False
        else:
            for i_spectrum, spectrum in enumerate(spectra1):
                if not numpy.array_equal(spectrum.mu, spectra2[i_spectrum].mu):
                    return False
            return True

    @property
    def n_spectrum(self):
        """return the number of spectra"""
        if self.__spectra is None:
            return 0
        else:
            return len(self.__spectra)

    def spectra_keys(self):
        """keys contained by the spectrum object (energy, mu, normalizedmu...)
        """
        if len(self.spectra) > 0:
            assert isinstance(self.spectra[0], Spectrum)
            return self.spectra[0].keys()

    @property
    def linked_h5_file(self):
        return self.__h5_file

    def link_to_h5(self, h5_file):
        """
        Associate a .h5 file to the XASObject. This can be used for storing
        process flow.
        
        :param h5_file: 
        :return: 
        """
        self.__h5_file = h5_file

    def has_linked_file(self):
        return self.__h5_file is not None

    def get_next_processing_index(self):
        self.__processing_index += 1
        return self.__processing_index

    def register_processing(self, process, data):
        """
        Register one process for the current xas object. This require to having
        link a h5file to this object

        :param process: process to be registred
        :type: :class:`.Process`
        :param data: result of the processing. If there is more than one
                     result then a dictionary with the key under which result
                     should be saved and a numpy.ndarray
        :type: Union[numpy.ndarray,dict]
        """
        est.io.write_xas_proc(self.linked_h5_file, entry=self.__entry_name,
                              processing_order=self.get_next_processing_index(),
                              process=process, data=data, overwrite=True)

    def get_process_flow(self):
        """
        
        :return: the dict of process information
        :rtype: dict
        """
        if not self.linked_h5_file:
            _logger.warning('process flow is store in the linked .h5 file. If'
                            'no link is defined then this information is not'
                            'stored')
            return {}
        else:
            recognized_process = xas.io.get_xasproc(self.linked_h5_file,
                                                    entry=self.__entry_name)
            know_process = ('pymca_normalization', 'pymca_exafs', 'pymca_ft',
                            'pymca_k_weight', 'larch_autobk', 'larch_mback',
                            'larch_mback_norm', 'larch_pre_edge', 'larch_xftf')

            def filter_recognized_process(process_list):
                res = []
                for process_ in process_list:
                    if 'program' in process_.keys() and process_['program'] in know_process:
                        res.append(process_)
                return res
            recognized_process = filter_recognized_process(recognized_process)

            def get_ordered_process(process_list):
                res = {}
                for process_ in process_list:
                    if not 'processing_order' in process_:
                        _logger.warning('one processing has not processing order: ' + process_['program'])
                    else:
                        processing_order = int(process_['processing_order'])
                        res[processing_order] = process_
                return res

            return get_ordered_process(recognized_process)

    def clean_process_flow(self):
        """
        Remove existing process flow
        """
        if not self.linked_h5_file:
            _logger.warning('process flow is store in the linked .h5 file. If'
                            'no link is defined then this information is not'
                            'stored')
        else:
            process_flow = self.get_process_flow()
            with h5py.File(self.linked_h5_file) as h5f:
                for index, process_ in process_flow.items():
                    del h5f[process_['_h5py_path']]

    def copy_process_flow_to(self, h5_file_target):
        """
        copy all the recognized process from self.__h5_file to h5_file_target

        :param str h5_file_target: path to the targeted file. Should be an 
                                   existing hdf5 file.
        """
        assert os.path.exists(h5_file_target)
        assert h5py.is_hdf5(h5_file_target)

        flow = self.get_process_flow()
        entry = self.entry
        with h5py.File(self.__h5_file) as source_hdf:
            with h5py.File(h5_file_target) as target_hdf:
                target_entry = target_hdf.require_group(entry)
                def remove_entry_prefix(name):
                    return name.replace('/'+entry+'/', '', 1)
                for process_id, process in flow.items():
                    process_path = process['_h5py_path']
                    dst_path = remove_entry_prefix(name=process_path)
                    target_entry.copy(source=source_hdf[process_path],
                                      dest=dst_path)


# TODO: add the spectra class. Would speed up and simplify stuff probably
class Spectra(object):
    pass


class Spectrum(_Spectrum_Base):
    """
    Core object to be used to store larch and pymca results.

    Larch is using 'Group' to store the result and add menbers to this group
    according to the different treatment.
    Pymca is using a dictionary to store the result.

    This class as to adpat to the both behavior and the different naming
    convention as well.

    :param numpy.ndarray (1D) energy: beam energy
    :param numpy.ndarray (1D) mu: beam absorption
    """
    _MU_KEY = 'Mu'

    _ENERGY_KEY = 'Energy'

    _NORMALIZED_MU_KEY = 'NormalizedMu'

    _NORMALIZED_ENERGY_KEY = 'NormalizedEnergy'

    _NORMALIZED_SIGNAL_KEY = 'NormalizedSignal'

    _EXAFS_SIGNAL_KEY = 'EXAFSSignal'

    _EXAFS_KVALUES_KEY = 'EXAFSKValues'

    _FT_KEY = 'FT'

    _EDGE_KEY = 'Edge'

    _NORMALIZED_BACKGROUND_KEY = 'NormalizedBackground'

    def __init__(self, energy=None, mu=None):
        _Spectrum_Base.__init__(self)
        if energy is not None:
            assert isinstance(energy, numpy.ndarray)

        # properties
        self.energy = energy
        self.mu = mu
        self.__chi = None
        self.__k_values = None
        self.__normalized_mu = None
        self.__normalized_energy = None
        self.__pre_edge = None
        self.__post_edge = None
        # unable to create a property top level since larch is using
        # getattr(group, 'edge', None). Or edge should be initialized to
        # something != None that we don't want
        self.__e0 = None
        self.__other_parameters = {}
        self.ft = {}

        self.__key_mapper = {
            self._MU_KEY: self.__class__.mu,
            self._ENERGY_KEY: self.__class__.energy,
            self._NORMALIZED_MU_KEY: self.__class__.normalized_mu,
            self._NORMALIZED_ENERGY_KEY: self.__class__.normalized_energy,
            self._NORMALIZED_SIGNAL_KEY: self.__class__.post_edge,
            self._NORMALIZED_BACKGROUND_KEY: self.__class__.pre_edge,
            self._FT_KEY: self.__class__.ft,
            self._EDGE_KEY: self.__class__.e0,
            self._EXAFS_SIGNAL_KEY: self.__class__.chi,
            self._EXAFS_KVALUES_KEY: self.__class__.k,
        }

    @property
    def energy(self):
        return self.__energy

    @energy.setter
    def energy(self, energy):
        assert isinstance(energy, numpy.ndarray) or energy is None
        self.__energy = energy

    @property
    def mu(self):
        return self.__mu

    @mu.setter
    def mu(self, mu):
        assert isinstance(mu, numpy.ndarray) or mu is None
        self.__mu = mu

    @property
    def chi(self):
        return self.__chi

    @chi.setter
    def chi(self, chi):
        self.__chi = chi

    @property
    def k(self):
        return self.__k_values

    @k.setter
    def k(self, k):
        self.__k_values = k

    @property
    def normalized_mu(self):
        return self.__normalized_mu

    @normalized_mu.setter
    def normalized_mu(self, mu):
        assert isinstance(mu, numpy.ndarray) or mu is None
        self.__normalized_mu = mu

    @property
    def norm(self):
        # this alias is needed for larch
        return self.__normalized_mu

    @norm.setter
    def norm(self, value):
        # this alias is needed for larch
        self.__normalized_mu = value

    @property
    def normalized_energy(self):
        return self.__normalized_energy

    @normalized_energy.setter
    def normalized_energy(self, energy):
        assert isinstance(energy, numpy.ndarray) or energy is None
        self.__normalized_energy = energy

    @property
    def pre_edge(self):
        return self.__pre_edge

    @pre_edge.setter
    def pre_edge(self, value):
        self.__pre_edge = value

    @property
    def _edge(self):
        return self.edge

    @_edge.setter
    def _edge(self, value):
        self.edge = value

    @property
    def post_edge(self):
        return self.__post_edge

    @post_edge.setter
    def post_edge(self, value):
        self.__post_edge = value

    @property
    def e0(self):
        return self.__e0

    @e0.setter
    def e0(self, e0):
        self.__e0 = e0

    @property
    def ft(self):
        return self.__ft

    @ft.setter
    def ft(self, ft):
        if isinstance(ft, _FT):
            self.__ft = ft
        else:
            self.__ft = _FT(ddict=ft)

    @property
    def shape(self):
        _energy_len = 0
        if self.__energy is not None:
            _energy_len = len(self.__energy)
        _mu_len = 0
        if self.__mu is not None:
            _mu_len = len(self.__mu)

        return (_energy_len, _mu_len)

    def extra_keys(self):
        return self.__other_parameters.keys()

    def __getitem__(self, key):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            return self.__key_mapper[key].fget(self)
        else:
            return self.__other_parameters[key]

    def __setitem__(self, key, value):
        """Need for pymca compatibility"""
        assert key
        if key in self.__key_mapper:
            self.__key_mapper[key].fset(self, value)
        else:
            self.__other_parameters[key] = value

    def __contains__(self, item):
        return item in self.__key_mapper or item in self.__other_parameters

    def load_frm_dict(self, ddict):
        assert isinstance(ddict, dict)
        larch_ddict = {}
        if '_larch_grp_members' in ddict:
            if _has_larch is False:
                _logger.warning('larch is not installed but the Spectrum '
                                'requires it, won\'t be able to load data '
                                'relative to larch')
            for key in ddict['_larch_grp_members']:
                larch_ddict[key] = ddict[key]
                del ddict[key]

        for key, value in ddict.items():
            self[key] = value
        for key, value in larch_ddict.items():
            setattr(self, key, value)
        return self

    def update(self, spectrum):
        assert isinstance(spectrum, Spectrum)
        for key in spectrum:
            self[key] = spectrum[key]

    @staticmethod
    def from_dict(ddict):
        spectrum = Spectrum()
        return spectrum.load_frm_dict(ddict=ddict)

    def to_dict(self):
        res = {
            self._MU_KEY: self.mu,
            self._ENERGY_KEY: self.energy,
            self._FT_KEY: self.ft.to_dict(),
            self._NORMALIZED_MU_KEY: self.normalized_mu,
            self._NORMALIZED_ENERGY_KEY: self.normalized_energy,
            self._NORMALIZED_SIGNAL_KEY: self.post_edge,
            self._NORMALIZED_BACKGROUND_KEY: self.pre_edge,
            self._EDGE_KEY: self.e0,
            self._EXAFS_SIGNAL_KEY: self.chi,
            self._EXAFS_KVALUES_KEY: self.k,
        }
        if _has_larch:
            res.update(self._getLarchGroupMenbers())
        res.update(self.__other_parameters)
        return res

    def _getLarchGroupMenbers(self):
        """Return larch group specific menbers"""
        assert _has_larch is True
        res = {}
        for key in self._members().keys():
            res[key] = getattr(self, key)
        res['_larch_grp_members'] = self._members().keys()
        return res

    def __str__(self):
        def add_info(str_, attr):
            assert hasattr(self, attr)
            sub_str = '- ' + attr + ': ' + str(getattr(self, attr)) + '\n'
            return (str_ + sub_str)
        main_info = ""
        for info in ('energy', 'mu', 'normalized_mu', 'normalized_signal', 'normalized_energy'):
            main_info = add_info(str_=main_info, attr=info)

        def add_third_info(str_, key):
            sub_str = ('- ' + key + ': ' + str(self[key])) + '\n'
            return str_ + sub_str
        for key in self.__other_parameters:
            main_info = add_third_info(str_=main_info, key=key)
        return main_info

    def update(self, obj):
        """
        Update the contained values from the given obj.

        :param obj:
        :type obj: Union[XASObject, dict]
        """
        if isinstance(obj, Spectrum):
            _obj = obj.to_dict()
        else:
            _obj = obj
        assert isinstance(_obj, dict)
        for key, value in _obj.items():
            self[key] = value

    def get_missing_keys(self, keys):
        """Return missing keys on the spectrum"""
        missing = []
        for key in keys:
            if key not in self:
                missing.append(key)
        if len(missing) is 0:
            return None
        else:
            return missing

    def keys(self):
        keys = list(self.__other_parameters.keys())
        keys += list(self.__key_mapper.keys())
        return keys

    def copy(self):
        return Spectrum().load_frm_dict(self.to_dict())


class _FT(object):

    _RADIUS_KEY = 'FTRadius'

    _INTENSITY_KEY = 'FTIntensity'

    _IMAGINARY_KEY = 'FTImaginary'

    def __init__(self, ddict):
        self.__radius = None
        self.__intensity = None
        self.__imaginary = None
        self.__other_parameters = {}

        self.__key_mapper = {
            self._RADIUS_KEY: self.__class__.radius,
            self._INTENSITY_KEY: self.__class__.intensity,
            self._IMAGINARY_KEY: self.__class__.imaginary,
        }

        if ddict is not None:
            for key, values in ddict.items():
                self[key] = values

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    def radius(self, radius):
        self.__radius = radius

    @property
    def intensity(self):
        return self.__intensity

    @intensity.setter
    def intensity(self, intensity):
        self.__intensity = intensity

    @property
    def imaginary(self):
        return self.__imaginary

    @imaginary.setter
    def imaginary(self, imaginery):
        self.__imaginary = imaginery

    def __getitem__(self, key):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            return self.__key_mapper[key].fget(self)
        else:
            return self.__other_parameters[key]

    def __setitem__(self, key, value):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            self.__key_mapper[key].fset(self, value)
        else:
            self.__other_parameters[key] = value

    def __contains__(self, item):
        return item in self.__key_mapper or item in self.__other_parameters

    def to_dict(self):
        res = {
            self._RADIUS_KEY: self.radius,
            self._INTENSITY_KEY: self.intensity,
            self._IMAGINARY_KEY: self.imaginary,
        }
        res.update(self.__other_parameters)
        return res

    def get_missing_keys(self, keys):
        """Return missing keys on the spectrum"""
        missing = []
        for key in keys:
            if key not in self:
                missing.append(key)
        if len(missing) is 0:
            return None
        else:
            return missing


class Sample(object):
    """Description of the sample. Needed for writing valid nx file"""
    def __init__(self, name='undefined sample', description=None):
        self.name = name
        self.description = description



import argparse
import sys
from ..pushworkflow.scheme.parser import scheme_load
import logging
try:
    import h5py
    has_h5py = True
except:
    has_h5py = False

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(__name__)


def exec_(scheme, input_=None, output_=None):
    def manage_input():
        if input_:
            # monkey patch the input file for start nodes if an input is given
            for node in scheme.nodes:
                if node.properties and '_input_file_setting' in node.properties:
                    node.properties['_input_file_setting'] = input_
    manage_input()

    def manage_output():
        if output_:
            # monkey patch the input file for start nodes if an input is given
            for node in scheme.nodes:
                if node.properties and '_output_file_setting' in node.properties:
                    node.properties['_output_file_setting'] = output_

    # TODO: if the input file contains configuration, then overwrite the node
    # properties...
    manage_output()

    if len(scheme._start_actor.listDownStreamActor) is 0:
        _logger.warning('no downstream node defined')
        return None
    scheme._start_actor.trigger(input_)
    scheme._end_actor.join()
    return scheme._end_actor.outData


def getinputinfo():
    return "xas process workflow_file_desc.ows [input]"


def getInputFrmHdf5(file_path):
    raise NotImplementedError()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'workflow_file',
        help='Workflow file providing the workflow description (.ows, .xml)')
    parser.add_argument("-i", dest="input_", default=None,
                        help="Input of the workflow. require at most one "
                             "instance of XASInputOW")
    parser.add_argument("-o", dest="output_", default=None,
                        help="Output file of the workflow. Require at most one "
                             "instance of XASOutputOW")
    options = parser.parse_args(argv[1:])

    scheme = scheme_load(_file=options.workflow_file)
    exec_(scheme, options.input_, options.output_)

if __name__ == "__main__":
    main(sys.argv)



from .scheme.node import Node, WorkflowException, _callback_info
import multiprocessing
import importlib
from importlib.machinery import SourceFileLoader
import logging

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
logger = logging.getLogger()


def get_handler(obj, data_name):
    """Return the name of the handler for this data_name"""
    def input_list_to_dict(inputs):
        res = {}
        # link each link name to the appropriate handler
        for input_ in inputs:
            res[input_[0]] = input_[2]
        return res
    input_dict = input_list_to_dict(obj.inputs)
    if data_name in input_dict:
        return input_dict[data_name]
    else:
        raise ValueError('link with name', data_name, 'is not managed by the object')


def _createCallback(name):
    sname = name.rsplit('.')
    if not (len(sname) > 1):
        raise ValueError('Invalid name')
    class_name = sname[-1]
    del sname[-1]
    module_name = '.'.join(sname)
    try:
        m = importlib.import_module(module_name)
    except ImportError:
        # try if given from file name
        del sname[-1]
        class_name = name.rsplit('.')[-1]
        m = SourceFileLoader(module_name, module_name).load_module()
    return getattr(m, class_name)()


def exec_process(callback_info, input_, properties):
    logger.debug('processing', str(callback_info), 'with input', input_, 'and',
                 properties, 'as properties')

    if type(input_) is tuple:
        data_name, data = input_
    else:
        data_name = None
        data = input_
    callback_info = _callback_info(callback=callback_info.callback,
                                   handler=callback_info.handler,
                                   need_instanciation=callback_info.need_instanciation)
    if callback_info.need_instanciation:
        import inspect
        if inspect.isclass(callback_info.callback):
            callback = callback_info.callback()
        else:
            callback = _createCallback(callback_info.callback)
    else:
        callback = callback_info.callback
    if hasattr(callback, 'setProperties'):
        callback.setProperties(properties)

    if data_name:
        # if the link specify a name, then we should be able to have several
        # handler
        handler = get_handler(callback, data_name)
    else:
        if not callable(callback):
            handler = 'process'
        else:
            handler = None

    # pick the default process function is no handler and object not callable
    if handler is not None:
        if not hasattr(callback, handler):
            raise ValueError('handler function', handler, 'does not exist for', callback)
        out = getattr(callback, handler)(data)
    else:
        out = callback(data)
    if hasattr(out, 'to_dict'):
        return out.to_dict()
    else:
        return out


class AsyncFactory:
    def __init__(self, actor, callback_info=None):
        self.actor = actor
        self.pool = multiprocessing.Pool()
        self.callback_info = callback_info

    def call(self, *args, **kwargs):
        logger.debug('args={0}, kwargs={1}'.format(args, kwargs))
        if self.callback_info is None:
            self.pool.apply_async(exec_process, args, kwargs)
        else:
            self.pool.apply_async(exec_process, args, kwargs, self.callback_info)


class ActorWrapper(object):
    def __init__(self, actor):
        assert isinstance(actor, Node)
        self.actor = actor

    def run(self, inData):

        outData = self.actor.run(inData)
        if isinstance(outData, WorkflowException):
            return outData
        else:
            inData.update(outData)
            return inData


class ActorFactory(object):
    def __init__(self, name, errorHandler=None):
        self.name = name
        self.errorHandler = errorHandler
        self._a = None
        self.listDownStreamActor = []
        self.outData = None
        self.actorWrapper = ActorWrapper(name)
        self.af = None
        self.lock = multiprocessing.Lock()

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def _trigger_process(self, inData):
        """Trigger one process"""
        logger.debug('In trigger {0}, inData = {1}'.format(self.name, inData))
        if isinstance(inData, WorkflowException):
            logger.error(
                    'Error from previous actor! Not running actor {0}'.format(
                            self.name))
            if self.errorHandler is not None:
                workflowException = inData
                oldInData = workflowException.data
                exceptionDict = {
                    'errorMessage': workflowException.errorMessage,
                    'traceBack': workflowException.traceBack,
                }
                oldInData['WorkflowException'] = exceptionDict
                self.errorHandler.triggerOnError(oldInData)
        elif len(self.listDownStreamActor) == 0:
            # Run without any callback
            self.af = AsyncFactory(self.actorWrapper.actor.callback)
            self.af.call(self.actorWrapper.actor.callback,
                         inData,
                         self.actorWrapper.actor.properties)
        else:
            # Run with callback
            self.af = AsyncFactory(self.actorWrapper.actor.callback,
                                   self.triggerDownStreamActor)
            self.af.call(self.actorWrapper.actor.callback,
                         inData,
                         self.actorWrapper.actor.properties)

    def trigger(self, data):
        if hasattr(data, 'to_dict'):
            _data = data.to_dict()
        else:
            _data = data
        if _data and 'sig_type' in _data:
            self._trigger_signal(data=_data)
        else:
            properties = self.actorWrapper.actor.properties
            if _data is None and '_scanIDs' in properties:
                inDataList = properties['_scanIDs']
            else:
                inDataList = [_data, ]

            for inData in inDataList:
                self._trigger_process(inData=inData)

            if self.actorWrapper.actor.need_stop_join:
                self._trigger_stop(len(inDataList))

    def _trigger_stop(self, nProcess):
        """Signal that this process is finished and the downstream nodes
        Can stop once they processed nProcess"""
        self._trigger_signal(data={'sig_type': 'stop', 'n_process': nProcess})

    def _trigger_signal(self, data):
        if len(self.listDownStreamActor) == 0:
            return
        else:
            self.triggerDownStreamActor(inData=data)

    def triggerDownStreamActor(self, inData=None):
        if inData is None:
            inData = {}
        for downStreamActor in self.listDownStreamActor:
            logger.debug(
                'In trigger {0}, triggering actor {1}, inData={2}'.format(
                    self.name, downStreamActor.name, inData))
            downStreamActor.trigger(inData)


class StartActor(object):
    def __init__(self, name='Start actor'):
        self.name = name
        self.listDownStreamActor = []

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        for actor in self.listDownStreamActor:
            actor.trigger(inData)


class StopActor(object):
    def __init__(self, name='Stop actor'):
        self.name = name
        self.lock = multiprocessing.Lock()
        self.lock.acquire()
        self.outData = None

    def trigger(self, inData):
        logger.debug('In trigger {0}, inData = {1}'.format(self.name, inData))
        self.outData = inData
        self.lock.release()

    def join(self, timeout=7200):
        self.lock.acquire(timeout=timeout)


class Join(object):
    def __init__(self, name, numberOfConnections):
        self.name = name
        self.numberOfConnections = numberOfConnections
        self.listInData = []
        self.listDownStreamActor = []

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        self.listInData.append(inData)
        if len(self.listInData) == self.numberOfConnections:
            newInData = {}
            for data in self.listInData:
                newInData.update(data)
            for actor in self.listDownStreamActor:
                actor.trigger(newInData)


class JoinUntilStopSignal(object):
    def __init__(self, name):
        self.name = name
        self.listInData = []
        self.listDownStreamActor = []
        self._nprocess_received = 0
        self._nprocess_waited = 0
        self._can_stop = False

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        print('***JoinUntilStopSignal***', inData)

        if type(inData) is dict and 'sig_type' in inData and inData['sig_type'] == 'stop':
            self._can_stop = True
            self._nprocess_waited = inData['n_process']
        else:
            self._nprocess_received += 1

        self.listInData.append(inData)
        print('_can_stop', self._can_stop, ', ', self._nprocess_waited <= self._nprocess_received)

        if self._can_stop and self._nprocess_waited <= self._nprocess_received:
            newInData = {}
            for data in self.listInData:
                newInData.update(data)
            for actor in self.listDownStreamActor:
                actor.trigger(newInData)


class RouterActor(object):
    def __init__(self, parent=None, name='Router', itemName=None, listPort=[]):
        self.parent = parent
        self.name = name
        self.itemName = itemName
        self.listPort = listPort
        self.dictValues = {}

    def connect(self, actor, expectedValue='other'):
        if expectedValue != 'other' and not expectedValue in self.listPort:
            raise RuntimeError(
                'Port {0} not defined for router actor {1}!'.format(
                    expectedValue, self.name))
        self.dictValues[expectedValue] = actor

    def trigger(self, inData):
        value = inData[self.itemName]
        if value in self.dictValues:
            self.dictValues[value].trigger(inData=inData)
        else:
            self.dictValues['other'].trigger(inData=inData)


class Port(object):
    def __init__(self, parent, name):
        self.name = parent.name + '.' + name
        self.parent = parent
        self.listActor = []
        self.inPortTrigger = None

    def connect(self, actor):
        logger.debug(
            "Connect {0} -> actorName {1}".format(self.name, actor.name))
        self.listActor.append(actor)

    def setTrigger(self, trigger):
        self.inPortTrigger = trigger

    def trigger(self, inData):
        logger.debug("In {0} trigger".format(self.name))
        if len(self.listActor) > 0:
            for actor in self.listActor:
                logger.debug(
                    "In trigger {0} -> actorName {1}".format(self.parent.name,
                                                             actor.name))
                actor.trigger(inData)
        if self.inPortTrigger is not None:
            logger.debug(
                "In {0} trigger, trigger = {1}".format(self.parent.name,
                                                       self.inPortTrigger))
            self.inPortTrigger(inData)


class SubModel(object):
    def __init__(self, errorHandler=None, name=None, portNames=['In', 'Out']):
        self.name = name
        self.errorHandler = errorHandler
        self.dictPort = {}
        self.listOnErrorActor = []
        for portName in portNames:
            self.dictPort[portName] = Port(self, portName)

    def getPort(self, portName):
        logger.debug(
            "In {0} getPort, portName = {1}".format(self.name, portName))
        return self.dictPort[portName]

    def connect(self, actor, portName='Out'):
        logger.debug("In {0} connect, portName = {2} -> actorName = {1}".format(
            self.name, actor.name, portName))
        self.dictPort[portName].connect(actor)

    def connectOnError(self, actor):
        logger.debug(
            "In connectOnError in subModule {0}, actor name {1}".format(
                self.name, actor.name))
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, inData):
        for onErrorActor in self.listOnErrorActor:
            logger.debug(
                "In triggerOnError in subModule {0}, trigger actor {1}, inData = {2}".format(
                    self.name, onErrorActor.name, args[0]))
            onErrorActor.trigger(inData)
        if self.errorHandler is not None:
            self.errorHandler.triggerOnError(inData)


class ErrorHandler(object):
    def __init__(self, name='Start actor'):
        self.name = name
        self.listDownStreamActor = []

    def connect(self, actor):
        self.listDownStreamActor.append(actor)

    def triggerOnError(self, inData):
        for actor in self.listDownStreamActor:
            actor.trigger(inData)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


import os
import unittest
from est.core.types import Spectrum, XASObject
try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.core.process.pymca.k_weight import pymca_k_weight

    from est.core.process.pymca.normalization import pymca_normalization
    from est.io.utils.pymca import read_spectrum


@unittest.skipIf(has_pymca is False, 'PyMca5 is not installed')
class TestKWeightSingleSpectrum(unittest.TestCase):
    """Make sure the process have valid io"""
    def setUp(self):
        self.config = {
            'SET_KWEIGHT': 2.0,
            'EXAFS': {"Knots": {"Values": (1, 2, 5), "Number": 3,
                                'Orders': [3, 3, 3]}},
        }

        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(data_file)

        self.spectrum = Spectrum(energy=energy, mu=mu)
        self.xas_obj = XASObject(energy=energy, spectra=(self.spectrum, ),
                                 configuration=self.config, dim1=1, dim2=1)

    def testWithXASObjAsInput(self):
        self.xas_obj = pymca_normalization(self.xas_obj)
        pymca_k_weight(xas_obj=self.xas_obj)

    def testWithDictAsInput(self):
        self.xas_obj = pymca_normalization(self.xas_obj)
        pymca_k_weight(xas_obj=self.xas_obj.to_dict())


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestKWeightSingleSpectrum,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/05/2017"

global next_link_free_id
next_link_free_id = 0


def get_next_link_free_id():
    global next_link_free_id
    _id = next_link_free_id
    next_link_free_id += 1
    return _id


class Link(object):
    """

    :param `.Node` source_node:
    :param `.Node` sink_node:
    :param str source_channel:
    :param str sink_channel:
    """

    def __init__(self, source_node, sink_node, source_channel, sink_channel,
                 id=None):
        self.id = id or get_next_link_free_id()
        if isinstance(source_node, int):
            self.source_node_id = source_node
        else:
            self.source_node_id = source_node.id

        if isinstance(sink_node, int):
            self.sink_node_id = sink_node
        else:
            self.sink_node_id = sink_node.id

        self.source_channel = source_channel
        self.sink_channel = sink_channel



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/05/2017"

import functools
import logging
import traceback
from collections import namedtuple
import inspect

_logger = logging.getLogger(__file__)

global next_node_free_idF
next_node_free_id = 0


def get_next_node_free_id():
    global next_node_free_id
    _id = next_node_free_id
    next_node_free_id += 1
    return _id


_callback_info = namedtuple("_callback_info",
                            ["callback", "handler", "need_instanciation"])


def trace_unhandled_exceptions(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            outData = func(*args, **kwargs)
        except Exception as e:
            _logger.exception(e)
            errorMessage = '{0}'.format(e)
            traceBack = traceback.format_exc()
            return WorkflowException(
                    msg=errorMessage,
                    traceBack=traceBack,
                    data=args[1]
            )
        return outData

    return wrapped_func


class Node(object):
    """
    Node in the `.Scheme`. Will be associated to a tomwer process.

    :param callback: pointer to a class or a function or str defining the
                     callback. If the callback is a class then the handler
                     should be defined or the class should have a default
                     'process' function that will be called by default.
    :param int id: unique id of the node.
    :param dict properties: properties of the node
    :param str luigi_task: luigi task associate to this node
    """

    need_stop_join = False
    """flag to stop the node only when receive the 'stop' signal"""

    def __init__(self, callback, id=None, properties=None,
                 error_handler=None):
        self.id = id or get_next_node_free_id()
        """int of the node id"""
        self.properties = properties or {}
        """dict of the node properties"""
        self.upstream_nodes = set()
        """Set of upstream nodes"""
        self.downstream_nodes = set()
        """Set of downstream nodes"""
        self.__process_instance = None
        """"""
        self.callback = callback
        """process instance"""
        self._error_handler = error_handler
        self.outData = None

    @property
    def callback(self):
        return self.__callback

    @callback.setter
    def callback(self, callback):
        need_instanciation = type(callback) is str or inspect.isclass(callback)
        self.__callback = _callback_info(callback=callback, handler=None,
                                         need_instanciation=need_instanciation)

    def isfinal(self):
        return len(self.downstream_nodes) is 0

    def isstart(self):
        return len(self.upstream_nodes) is 0


class WorkflowException(Exception):
    def __init__(self, traceBack="", data=None, msg=None):
        if data is None:
            data = {}
        super(WorkflowException, self).__init__(msg)
        self.errorMessage = msg
        self.data = data
        self.traceBack = traceBack



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/30/2019"


from silx.gui import qt
from est.gui.larch.utils import _OptionalQDoubleSpinBox, _OptionalQIntSpinBox


class _MPreEdgeParameters(qt.QWidget):
    """
    Widget for setting the configuration of the larch 'pre edge' process
    """
    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())

        # e0
        self._e0SB = _OptionalQDoubleSpinBox(parent=self)
        self._e0SB.setToolTip("edge energy, in eV.  If None, it will be "
                              "determined here.")
        self._e0SB.setMinimum(0.0)
        self._e0SB.setValue(2000.0)
        self._e0SB.setValue(None)
        self.layout().addRow(qt.QLabel('e0', parent=self), self._e0SB)
        # step
        self._stepSP = _OptionalQDoubleSpinBox(parent=self)
        self._stepSP.setToolTip('edge jump. If None, it will be determined')
        self._stepSP.setValue(None)
        self.layout().addRow(qt.QLabel('step', parent=self), self._stepSP)
        # pre1
        self._pre1SB = _OptionalQDoubleSpinBox(parent=self)
        self._pre1SB.setToolTip("low E range (relative to e0) for pre-edge "
                                "region")
        self._pre1SB.setValue(None)
        self._pre1SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addRow(qt.QLabel('pre1', parent=self), self._pre1SB)
        # pre2
        self._pre2SB = qt.QDoubleSpinBox(parent=self)
        self._pre2SB.setToolTip("high E range (relative to e0) for pre-edge "
                                "region")
        self._pre2SB.setMinimum(-99999)
        self._pre2SB.setMaximum(99999)
        self._pre2SB.setValue(-50)
        self._pre2SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addRow(qt.QLabel('pre2', parent=self), self._pre2SB)
        # norm1
        self._norm1SB = qt.QDoubleSpinBox(parent=self)
        self._norm1SB.setToolTip("low E range (relative to e0) for post-edge "
                                 "region")
        self._norm1SB.setContentsMargins(0, 0, 0, 0)
        self._norm1SB.setMinimum(-99999)
        self._norm1SB.setMaximum(99999)
        self._norm1SB.setValue(100)
        self.layout().addRow(qt.QLabel('norm1', parent=self), self._norm1SB)
        # norm2
        self._norm2SB = _OptionalQDoubleSpinBox(parent=self)
        self._norm2SB.setToolTip("high E range (relative to e0) for post-edge "
                                 "region")
        self._norm2SB.setValue(None)
        self._norm2SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addRow(qt.QLabel('norm2', parent=self), self._norm2SB)
        # nvict
        self._nvictSB = qt.QSpinBox(parent=self)
        self._nvictSB.setValue(0)
        self.layout().addRow(qt.QLabel('energy exponent', parent=self),
                             self._nvictSB)
        # nnorm
        self._nnorm =_OptionalQIntSpinBox(parent=self)
        self._nnorm.setRange(0, 3)
        self._nnorm.setValue(None)
        self._nnorm.setToolTip('degree of polynomial (ie, nnorm+1 coefficients '
                               'will be found) for post-edge normalization '
                               'curve. If unset nnorm will default to 2 in '
                               'norm2-norm1>400, to 1 if 100>norm2-norm1>300, '
                               'and to 0 in norm2-norm1<100.')
        self.layout().addRow(qt.QLabel('polynomial degree', parent=self),
                             self._nnorm)
        # make_flat
        self._makeFlatCB = qt.QCheckBox('', parent=self)
        self._makeFlatCB.setToolTip('to calculate flattened output')
        self._makeFlatCB.setChecked(True)
        self.layout().addRow(qt.QLabel('make flat', parent=self),
                             self._makeFlatCB)
        # emin_area
        self._eminAreaSB = _OptionalQDoubleSpinBox(parent=self)
        self._eminAreaSB.setMinimum(0.0)
        self._eminAreaSB.setValue(None)
        self._eminAreaSB.setToolTip('Energy threshold for area normalization.'
                                    'norm_area will be estimated so that the '
                                    'area between emin_area and norm2 is equal '
                                    'to (norm2-emin_area). By default emin_area'
                                    ' will be set to the *nominal* edge energy '
                                    'for the element and '
                                    'edge - 3*core_level_width')
        self.layout().addRow(qt.QLabel('e min area', parent=self),
                             self._eminAreaSB)

        # expose API
        self.getStep = self._stepSP.getValue
        self.setStep = self._stepSP.setValue
        self.getPre1 = self._pre1SB.getValue
        self.setPre1 = self._pre1SB.setValue
        self.getPre2 = self._pre2SB.value
        self.setPre2 = self._pre2SB.setValue
        self.getNorm1 = self._norm1SB.value
        self.setNorm1 = self._norm1SB.setValue
        self.getNorm2 = self._norm2SB.getValue
        self.setNorm2 = self._norm2SB.setValue
        self.getNNorm = self._nnorm.getValue
        self.setNNorm = self._nnorm.setValue
        self.getNVict = self._nvictSB.value
        self.setNVict = self._nvictSB.setValue
        self.isUsingMakeFlat = self._makeFlatCB.isChecked
        self.setMakeFlat = self._makeFlatCB.setChecked
        self.getEMinArea = self._eminAreaSB.getValue
        self.setEMinArea = self._eminAreaSB.setValue

        # connect signal / Slot
        self._stepSP.sigChanged.connect(self._valueChanged)
        self._e0SB.sigChanged.connect(self._valueChanged)
        self._pre1SB.sigChanged.connect(self._valueChanged)
        self._pre2SB.editingFinished.connect(self._valueChanged)
        self._norm1SB.editingFinished.connect(self._valueChanged)
        self._norm2SB.sigChanged.connect(self._valueChanged)
        self._nnorm.sigChanged.connect(self._valueChanged)
        self._nvictSB.editingFinished.connect(self._valueChanged)
        self._makeFlatCB.toggled.connect(self._valueChanged)
        self._eminAreaSB.sigChanged.connect(self._valueChanged)

        # avoid several emission of the sigChanged signal
        self._lastParameters = None

    def _valueChanged(self, *arg, **kwargs):
        current_values = self.getParameters()
        if current_values != self._lastParameters:
            self._lastParameters = current_values
            self.sigChanged.emit()

    def getParameters(self):
        return {
            'e0': self.getE0(),
            'step': self.getStep(),
            'pre1': self.getPre1(),
            'pre2': self.getPre2(),
            'norm1': self.getNorm1(),
            'norm2': self.getNorm2(),
            'nnorm': self.getNNorm(),
            'nvict': self.getNVict(),
            'make_flat': self.isUsingMakeFlat(),
            'emin_area': self.getEMinArea(),
        }

    def setParameters(self, parameters):
        assert isinstance(parameters, dict)
        for key, value in parameters.items():
            if key == 'step':
                self.setStep(value)
            elif key == 'e0':
                self.setE0(value)
            elif key == 'pre1':
                self.setPre1(value)
            elif key == 'pre2':
                self.setPre2(value)
            elif key == 'norm1':
                self.setNorm1(value)
            elif key == 'norm2':
                self.setNorm2(value)
            elif key == 'nnorm':
                self.setNNorm(value)
            elif key == 'nvict':
                self.setNVict(value)
            elif key == 'make_flat':
                self.setMakeFlat(value)
            elif key == 'emin_area':
                self.setEMinArea(value)

    def getE0(self):
        return self._e0SB.getValue()

    def setE0(self, value):
        self._e0SB.setValue(value=value)


if __name__ == '__main__':
    app = qt.QApplication([])
    widget = _MPreEdgeParameters()
    widget.show()
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

from PyMca5.PyMcaIO import specfilewrapper as specfile


def read_spectrum(spec_file):
    """
    
    :param spec_file: path to the spec file containing the spectra definition
    :return: (energy, mu)
    :rtype: tuple
    """

    scan = specfile.Specfile(spec_file)[0]
    data = scan.data()

    if data.shape[0] == 2:
        energy = data[0, :]
        mu = data[1, :]
    else:
        energy = None
        mu = None
        labels = scan.alllabels()
        i = 0
        for label in labels:
            if label.lower() == "energy":
                energy = data[i, :]
            elif label.lower() in ["counts", "mu", "absorption"]:
                mu = data[i, :]
            i = i + 1
        if (energy is None) or (mu is None):
            if len(labels) == 3:
                if labels[0].lower() == "point":
                    energy = data[1, :]
                    mu = data[2, :]
                else:
                    energy = data[0, :]
                    mu = data[1, :]
            else:
                energy = data[0, :]
                mu = data[1, :]
    return energy, mu



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""simple helper function to link io module and XASObject"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/16/2019"


from est.io import read_xas, write_xas, get_xasproc
from est.core.types import XASObject
from silx.io.url import DataUrl
import h5py
import logging

_logger = logging.getLogger(__name__)

DEFAULT_SPECTRA_PATH = '/data/NXdata/data'

DEFAULT_CHANNEL_PATH = '/data/NXdata/Channel'

DEFAULT_CONF_PATH = '/configuration'


def read(spectra_url, channel_url, config_url=None):
    """
    
    :param DataUrl spectra_url: data url to the spectra
    :param DataUrl channel_url: data url to the channel / energy
    :param DataUrl config_url: data url to the process configuration
    :return: 
    :rtype: XASObject
    """
    reader = XASReader()
    return reader.read_frm_url(spectra_url=spectra_url, channel_url=channel_url,
                               config_url=config_url)


def read_frm_file(file_path):
    """
    
    :param str file_path: path to the file containing the spectra. Must ba a
                          .dat file that pymca can handle or a .h5py with
                          default path
    :rtype: XASObject
    """
    reader = XASReader()
    return reader.read_from_file(file_path=file_path)


class XASReader(object):
    """Simple reader of a xas file"""
    def read_frm_url(self, spectra_url, channel_url, config_url=None):
        sp, en, conf = read_xas(spectra_url=spectra_url,
                                channel_url=channel_url,
                                config_url=config_url)
        return XASObject(spectra=sp, energy=en, configuration=conf)

    def read_from_file(self, file_path):
        """

        :param str file_path: 
        :return: `.XASObject`
        """
        if file_path.endswith('.dat'):
            return self.read_frm_url(
                    spectra_url=DataUrl(file_path=file_path,
                                        scheme='PyMca'),
                    channel_url=DataUrl(file_path=file_path,
                                        scheme='PyMca'))
        elif file_path.endswith('.xmu'):
            return self.read_frm_url(
                    spectra_url=DataUrl(file_path=file_path,
                                        scheme='larch'),
                    channel_url=DataUrl(file_path=file_path,
                                        scheme='larch'))
        elif h5py.is_hdf5(file_path):
            return self.read_frm_url(
                    spectra_url=DataUrl(file_path=file_path,
                                        scheme='silx',
                                        data_path=DEFAULT_SPECTRA_PATH),
                    channel_url=DataUrl(file_path=file_path,
                                        scheme='silx',
                                        data_path=DEFAULT_CHANNEL_PATH),
                    config_url=DataUrl(file_path=file_path,
                                       scheme='silx',
                                       data_path='configuration'))
        else:
            raise ValueError('file type not managed, unable to load')

    __call__ = read_from_file



class XASWriter(object):
    """
    class to write the output file. In this case we need a class in order to
    setup the output file before
    """
    def __init__(self):
        self._output_file = None

    @property
    def output_file(self):
        return self._output_file

    @output_file.setter
    def output_file(self, file_):
        self._output_file = file_

    def setProperties(self, properties):
        if '_output_file_setting' in properties:
            self._output_file = properties['_output_file_setting']

    def dump_xas(self, xas_obj, write_process=True):
        """
        write the XASObject into a hdf5 file.

        :param xas_obj: object to be stored
        :type: Union[:class:`.XASObject`,dict]
        :param bool write_process: if True then store the process flow in the
            same file.
        """
        if isinstance(xas_obj, dict):
            _xas_obj = XASObject.from_dict(xas_obj)
        else:
            _xas_obj = xas_obj

        if not self._output_file:
            _logger.warning('no output file defined, please give path to the'
                            'output file')
            self._output_file = input()

        _logger.info(('dump xas obj to', self._output_file))

        # write raw data
        write_xas(h5_file=self._output_file, energy=_xas_obj.energy,
                  mu=_xas_obj.absorbed_beam(), entry=_xas_obj.entry)

        if write_process is True:
            if len(get_xasproc(self._output_file, entry=_xas_obj.entry)) > 0:
                _logger.warning('removing previous process registred. They are no '
                                'more accurate')
                _xas_obj.clean_process_flow()

            # write process flow
            _xas_obj.copy_process_flow_to(self._output_file)

    __call__ = dump_xas



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides access to the full silx test suite.

It is possible to disable tests depending on Qt by setting
`silx.test.utils.test_options.WITH_QT_TEST = False`
It will skip all tests from :mod:`silx.test.gui`.
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/12/2019"

import logging
import unittest
logger = logging.getLogger(__name__)


def suite():
    from ..core import test as test_core
    from ..app import test as test_app
    from ..pushworkflowactors import test as test_workflow_cases

    test_suite = unittest.TestSuite()
    # test sx first cause qui tests load ipython module
    test_suite.addTest(test_core.suite())
    test_suite.addTest(test_workflow_cases.suite())
    test_suite.addTest(test_app.suite())

    return test_suite


def run_tests():
    """Run test complete test_suite"""
    runner = unittest.TextTestRunner()
    if not runner.run(suite()).wasSuccessful():
        print("Test suite failed")
        return 1
    else:
        print("Test suite succeeded")
        return 0



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import numpy
import os
import unittest
import tempfile
import h5py
from est.core.types import Spectrum, XASObject
from est.core.utils import spectra as spectra_utils
from est.core.io import read as read_xas
from silx.io.url import DataUrl
import silx.io.utils
try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.io.utils.pymca import read_spectrum


class TestSpectrum(unittest.TestCase):
    """Test the spectrum class"""
    @unittest.skipIf(has_pymca is False, 'PyMca5 is not installed')
    def test_from_dat(self):
        """check that we can create a Spectrum from a pymca .dat file"""
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(data_file)
        spectrum = Spectrum(energy=energy, mu=mu)
        self.assertTrue(spectrum.energy is not None)
        self.assertTrue(spectrum.mu is not None)

    def test_from_numpy_array(self):
        """check that we can create a Spectrum from numpy arrays"""
        energy = numpy.arange(10, 20)
        mu = numpy.arange(10)
        spectrum = Spectrum(energy=energy, mu=mu)
        numpy.testing.assert_array_equal(spectrum.energy, energy)
        numpy.testing.assert_array_equal(spectrum.mu, mu)
        mu_2 = numpy.arange(30, 40)
        spectrum['Mu'] = mu_2
        numpy.testing.assert_array_equal(spectrum.mu, mu_2)


class TestXASObject(unittest.TestCase):
    """test construction of XAS object from a single spectra"""
    @unittest.skipIf(has_pymca is False, 'PyMca5 is not installed')
    def test_create_from_single_spectrum(self):
        """check that we can create a XASObject from a pymca .dat file"""
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        spectrum = {}
        spectrum['Energy'], spectrum['Mu'] = read_spectrum(data_file)
        self.spectrum = Spectrum(energy=spectrum['Energy'], mu=spectrum['Mu'])
        self.configuration = {
            'FT': {'KWeight': 1},
            'EXAFS': {'EXAFSNormalized': numpy.array([1, 2, 3])},
        }
        obj = XASObject(spectra=(self.spectrum,),
                        energy=self.spectrum.energy,
                        configuration=self.configuration,
                        dim1=1,
                        dim2=1)
        self.assertEqual(obj.n_spectrum, 1)
        ddict = obj.to_dict()
        obj2 = XASObject.from_dict(ddict)
        self.assertEqual(obj2, obj)

    def test_create_from_several_spectrums(self):
        """check that we can create a XASObject from numpy arrays"""
        self.energy, self.spectra = spectra_utils.create_dataset(shape=(256, 20, 10))
        self.output_dir = tempfile.mkdtemp()
        spectra_path = '/data/NXdata/data'
        channel_path = '/data/NXdata/Channel'
        filename = os.path.join(self.output_dir, 'myfile.h5')
        with h5py.File(filename) as f:
            f[spectra_path] = self.spectra
            f[channel_path] = self.energy

        url_spectra = DataUrl(file_path=filename,
                              data_path=spectra_path,
                              scheme='silx')
        self.xas_obj = read_xas(spectra_url=url_spectra,
                                channel_url=DataUrl(file_path=filename,
                                                    data_path=channel_path,
                                                    scheme='silx')
                                )
        self.assertEqual(self.xas_obj.dim1, 20)
        self.assertEqual(self.xas_obj.dim2, 10)
        self.assertEqual(self.xas_obj.n_spectrum, 20*10)
        ddict = self.xas_obj.to_dict(with_process_details=False)
        original_spectra = silx.io.utils.get_data(DataUrl(file_path=filename,
                                                          data_path=spectra_path,
                                                          scheme='silx'))
        numpy.testing.assert_array_equal(original_spectra,  ddict['spectra'])
        obj2 = XASObject.from_dict(ddict)
        self.assertEqual(self.xas_obj.n_spectrum,  obj2.n_spectrum)
        self.assertEqual(obj2, self.xas_obj)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSpectrum, TestXASObject, ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



import argparse
import sys
import logging
from est.core.types import XASObject
from est.core.reprocessing import get_process_instance_frm_h5_desc
try:
    import h5py
    has_h5py = True
except:
    has_h5py = False

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(__name__)


def exec_(h5_file, entry='scan1', spectra_path=None, energy_path=None,
          configuration_path=None):

    try:
        xas_obj = XASObject.from_file(h5_file=h5_file,
                                      entry=entry or 'scan1',
                                      spectra_path=spectra_path or 'data/absorbed_beam',
                                      energy_path=energy_path or 'data/energy',
                                      configuration_path=configuration_path)
        xas_obj.link_to_h5(h5_file=h5_file)
    except Exception as e:
        _logger.error(e)
        xas_obj = None

    if not xas_obj:
        _logger.error('Unable to create an object from the given file and path')

    xas_processes = xas_obj.get_process_flow()
    if len(xas_processes) == 0:
        _logger.warning('no xas process found in the given h5 file')
        return
    xas_obj.clean_process_flow()
    for i_process, process in xas_processes.items():
        process_ins = get_process_instance_frm_h5_desc(process)
        xas_obj = process_ins(xas_obj)

    return xas_obj


def getinputinfo():
    return "xas reprocessing workflow_file.h5"


def getInputFrmHdf5(file_path):
    raise NotImplementedError()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'workflow_file',
        help='Workflow file providing the workflow description (.h5)')
    parser.add_argument("--spectra_path", dest="spectra_path", default=None,
                        help="path to the spectra data")
    parser.add_argument("--energy_path", dest="energy_path", default=None,
                        help="path to the energy data")
    parser.add_argument("--configuration_path", dest="configuration_path", default=None,
                        help="path to the process configuration")
    parser.add_argument("--entry", dest="entry to load",
                        default=None,
                        help="path to the process configuration")
    options = parser.parse_args(argv[1:])
    exec_(h5_file=options.workflow_file, spectra_path=options.spectra_path,
          energy_path=options.energy_path,
          configuration_path=options.configuration_path)


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "17/12/2018"


from xml.etree.ElementTree import TreeBuilder, Element, ElementTree
from collections import defaultdict
from itertools import count, chain
import json
import pprint
import base64
import pickle
import logging
from .node import Node
from ..workflow import ActorFactory, StartActor, StopActor, JoinUntilStopSignal

_logger = logging.getLogger(__name__)


class Scheme(object):
    """
    class to define a workflow scheme from nodes and links

    :param list nodes:
    :param list links:
    """
    def __init__(self, nodes=None, links=None):
        self.title = ''
        self.description = ''
        self.links = {}
        """keys are link ID, values are Link"""
        if links is not None:
            for link in links:
                self.links[link.id] = link
        self.nodes = nodes or []
        """list of nodes"""
        self.nodes_dict = {}
        """dict with node id as key and node as value"""
        for node in self.nodes:
            self.nodes_dict[node.id] = node

        if links is not None:
            self._update_nodes_from_links()

        self._actor_factory = {}
        for node in self.nodes:
            self._actor_factory[node] = ActorFactory(node)

        # deal with connect
        for node in self.nodes:
            actor_factory = self._actor_factory[node]
            for downstream_node in node.downstream_nodes:
                downstream_actor_factory = self._actor_factory[downstream_node]
                actor_factory.connect(downstream_actor_factory)

        # add start actor
        self._start_actor = StartActor()
        for node in self.start_nodes():
            actor_factory = self._actor_factory[node]
            self._start_actor.connect(actor_factory)

        def connect_finals_nodes(actor):
            # add end actor
            for node in self.finalsNodes():
                actor_factory = self._actor_factory[node]
                actor_factory.connect(actor)

        self._end_actor = StopActor()

        if self.has_final_join():
            self._join_actor = JoinUntilStopSignal('stop join')
            connect_finals_nodes(self._join_actor)
            self._join_actor.connect(self._end_actor)
        else:
            connect_finals_nodes(self._end_actor)

    def finalsNodes(self):
        """

        :return: list of final nodes (with no output) and which hasn't any
                 control node upstream
        """
        res = []
        for node in self.nodes:
            assert isinstance(node, Node)
            if node.isfinal():
                res.append(node)
        return res

    def start_nodes(self):
        res = []
        for node in self.nodes:
            assert isinstance(node, Node)
            if node.isstart():
                res.append(node)
        return res

    def endlessNodes(self):
        res = []
        for node in self.nodes:
            assert isinstance(node, Node)
            if node.endless is True:
                res.append(node)
        return res

    def save_to(self, output_file):
        """
        Save the scheme as an xml formated file to `stream`
        """
        tree = self.scheme_to_etree(data_format="literal")
        indent(tree.getroot(), 0)

        tree.write(output_file)

    def scheme_to_etree(self, data_format="literal", pickle_fallback=False):
        """
        Return an `xml.etree.ElementTree` representation of the `scheme.
        """
        builder = TreeBuilder(element_factory=Element)
        builder.start("scheme", {"version": "2.0",
                                 "title": self.title or "",
                                 "description": self.description or ""})

        # Nodes
        node_ids = defaultdict(count().__next__)
        builder.start("nodes", {})
        for node in self.nodes:  # type: SchemeNode
            attrs = {"id": node.id,
                     # "name": node.name,
                     "qualified_name": node._qualified_name,
                     # "project_name": node.project_name or "",
                     # "version": node.version or "",
                     # "title": node.title,
                     }

            if type(node) is not Node:
                attrs["scheme_node_type"] = "%s.%s" % (type(node).__name__,
                                                       type(node).__module__)
            builder.start("node", attrs)
            builder.end("node")

        builder.end("nodes")

        # Links
        link_ids = defaultdict(count().__next__)
        builder.start("links", {})
        for link in self.links:  # type: SchemeLink
            source = link.source_node_id
            sink = link.sink_node_id
            source_id = node_ids[source]
            sink_id = node_ids[sink]
            attrs = {"id": str(link_ids[link]),
                     "source_node_id": str(source_id),
                     "sink_node_id": str(sink_id),
                     "source_channel": link.source_channel,
                     "sink_channel": link.sink_channel,
                     "enabled": "true" if link.enabled else "false",
                     }
            builder.start("link", attrs)
            builder.end("link")

        builder.end("links")

        # Annotations
        annotation_ids = defaultdict(count().__next__)
        builder.start("thumbnail", {})
        builder.end("thumbnail")

        # Node properties/settings
        builder.start("node_properties", {})
        for node in self.nodes:
            data = None
            if node.properties:
                try:
                    data, format = dumps(node.properties, format=data_format,
                                         pickle_fallback=pickle_fallback)
                except Exception:
                    _logger.error("Error serializing properties for node %r",
                              node.title, exc_info=True)
                if data is not None:
                    builder.start("properties",
                                  {"node_id": str(node_ids[node]),
                                   "format": format})
                    builder.data(data)
                    builder.end("properties")

        builder.end("node_properties")

        builder.end("scheme")
        root = builder.close()
        tree = ElementTree(root)
        return tree

    def _update_nodes_from_links(self):
        """
        Update upstream and downstream nodes from links definition
        """
        self._clear_nodes_connections()
        for link_id, link in self.links.items():
            source_node = self.nodes_dict[link.source_node_id]
            sink_node = self.nodes_dict[link.sink_node_id]
            source_node.downstream_nodes.add(self.nodes_dict[link.sink_node_id])
            sink_node.upstream_nodes.add(self.nodes_dict[link.source_node_id])

    def _clear_nodes_connections(self):
        """
        clear for all nodes downstream and upstream nodes
        """
        for node in self.nodes:
            assert isinstance(node, Node)
            node.downstream_nodes = set()
            node.upstream_nodes = set()

    def has_final_join(self):
        """True if we need to send a 'end' signal before closing the workflow
        This is needed for DataList and DataWatcher
        """
        for node in self.nodes:
            if node.need_stop_join:
                return True
        return False


def contains_control_nodes(nodes_list):
    for _node in nodes_list:
        if _node.endless or contains_control_nodes(_node.upstream_nodes):
            return True
    return False



def indent(element, level=0, indent="\t"):
    """
    Indent an instance of a :class:`Element`. Based on
    (http://effbot.org/zone/element-lib.htm#prettyprint).

    """
    def empty(text):
        return not text or not text.strip()

    def indent_(element, level, last):
        child_count = len(element)

        if child_count:
            if empty(element.text):
                element.text = "\n" + indent * (level + 1)

            if empty(element.tail):
                element.tail = "\n" + indent * (level + (-1 if last else 0))

            for i, child in enumerate(element):
                indent_(child, level + 1, i == child_count - 1)

        else:
            if empty(element.tail):
                element.tail = "\n" + indent * (level + (-1 if last else 0))

    return indent_(element, level, True)



def dumps(obj, format="literal", prettyprint=False, pickle_fallback=False):
    """
    Serialize `obj` using `format` ('json' or 'literal') and return its
    string representation and the used serialization format ('literal',
    'json' or 'pickle').

    If `pickle_fallback` is True and the serialization with `format`
    fails object's pickle representation will be returned

    """
    if format == "literal":
        try:
            return (literal_dumps(obj, prettyprint=prettyprint, indent=1),
                    "literal")
        except (ValueError, TypeError) as ex:
            if not pickle_fallback:
                raise

            _logger.debug("Could not serialize to a literal string")

    elif format == "json":
        try:
            return (json.dumps(obj, indent=1 if prettyprint else None),
                    "json")
        except (ValueError, TypeError):
            if not pickle_fallback:
                raise

            _logger.debug("Could not serialize to a json string")

    elif format == "pickle":
        return base64.encodebytes(pickle.dumps(obj)).decode('ascii'), "pickle"

    else:
        raise ValueError("Unsupported format %r" % format)

    if pickle_fallback:
        _logger.warning("Using pickle fallback")
        return base64.encodebytes(pickle.dumps(obj)).decode('ascii'), "pickle"
    else:
        raise Exception("Something strange happened.")



# This is a subset of PyON serialization.
def literal_dumps(obj, prettyprint=False, indent=4):
    """
    Write obj into a string as a python literal.
    """
    memo = {}
    NoneType = type(None)

    def check(obj):
        if type(obj) in [int, float, bool, NoneType, str, bytes]:
            return True

        if id(obj) in memo:
            raise ValueError("{0} is a recursive structure".format(obj))

        memo[id(obj)] = obj

        if type(obj) in [list, tuple]:
            return all(map(check, obj))
        elif type(obj) is dict:
            return all(map(check, chain(iter(obj.keys()), iter(obj.values()))))
        else:
            raise TypeError("{0} can not be serialized as a python "
                             "literal".format(type(obj)))

    check(obj)

    if prettyprint:
        return pprint.pformat(obj, indent=indent)
    else:
        return repr(obj)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/12/2019"


import os
import unittest
import est.core.process.roi
import est.core.io
from est.app.process import exec_ as exec_workflow
from est.core.types import XASObject
from est.pushworkflow.scheme.link import Link
from est.pushworkflow.scheme.node import Node
from est.pushworkflow.scheme.scheme import Scheme
try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    import est.core.process.pymca.exafs
    import est.core.process.pymca.ft
    import est.core.process.pymca.k_weight
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    import est.core.process.pymca.normalization


@unittest.skipIf(has_pymca is False, 'PyMca5 is not installed')
class TestWorkflowFunctions(unittest.TestCase):
    """Test some processing workflow"""
    def setUp(self):
        unittest.TestCase.setUp(self)

        read_task = Node(callback=est.core.io.read_frm_file)
        normalization_task = Node(
            callback=est.core.process.pymca.normalization.pymca_normalization)
        exafs_task = Node(callback=est.core.process.pymca.exafs.pymca_exafs)
        k_weight_task = Node(callback=est.core.process.pymca.k_weight.pymca_k_weight)
        ft_task = Node(callback=est.core.process.pymca.ft.pymca_ft)

        nodes = (
        read_task, normalization_task, exafs_task, k_weight_task, ft_task)

        links = [
            Link(source_node=read_task, source_channel='xas_obj',
                 sink_node=normalization_task, sink_channel='xas_obj'),
            Link(source_node=normalization_task, source_channel='xas_obj',
                 sink_node=exafs_task, sink_channel='xas_obj'),
            Link(source_node=exafs_task, source_channel='xas_obj',
                 sink_node=k_weight_task, sink_channel='xas_obj'),
            Link(source_node=k_weight_task, source_channel='xas_obj',
                 sink_node=ft_task, sink_channel='xas_obj'),
        ]

        self.scheme = Scheme(nodes=nodes, links=links)

        self.data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")

    def testProcess(self):
        out = exec_workflow(scheme=self.scheme, input_=self.data_file)
        assert isinstance(out, dict)
        xas_obj_out = XASObject.from_dict(out)
        assert 'FT' in xas_obj_out.spectra[0]
        assert 'FTRadius' in xas_obj_out.spectra[0].ft


@unittest.skipIf(has_pymca is False, 'PyMca5 is not installed')
class TestWorkflowCallableClasses(unittest.TestCase):
    """Test some processing workflow"""
    def setUp(self):
        unittest.TestCase.setUp(self)

        read_task = Node(callback=est.core.io.XASReader)
        roi_task = Node(callback=est.core.process.roi.ROIProcess)

        normalization_task = Node(
            callback=est.core.process.pymca.normalization.PyMca_normalization)
        k_weight_task = Node(callback=est.core.process.pymca.k_weight.PyMca_k_weight)
        exafs_task = Node(callback=est.core.process.pymca.exafs.PyMca_exafs)
        ft_task = Node(callback=est.core.process.pymca.ft.PyMca_ft)

        nodes = (read_task, roi_task, normalization_task, k_weight_task,
                 exafs_task, ft_task)

        links = [
            Link(source_node=read_task, source_channel='spectra',
                 sink_node=roi_task, sink_channel='spectra'),
            Link(source_node=roi_task, source_channel='spectra',
                 sink_node=normalization_task, sink_channel='spectra'),
            Link(source_node=normalization_task, source_channel='spectra',
                 sink_node=exafs_task, sink_channel='spectra'),
            Link(source_node=exafs_task, source_channel='spectra',
                 sink_node=k_weight_task, sink_channel='spectra'),
            Link(source_node=k_weight_task, source_channel='spectra',
                 sink_node=ft_task, sink_channel='spectra'),
        ]

        self.scheme = Scheme(nodes=nodes, links=links)

        self.data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")

    def testProcess(self):
        out = exec_workflow(scheme=self.scheme, input_=self.data_file)
        assert isinstance(out, dict)
        xas_obj_out = XASObject.from_dict(out)
        assert 'FT' in xas_obj_out.spectra[0]
        assert 'FTRadius' in xas_obj_out.spectra[0].ft
        assert 'FT' in xas_obj_out.configuration
        assert 'Normalization' in xas_obj_out.configuration


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestWorkflowFunctions, TestWorkflowCallableClasses):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "03/07/2019"


from silx.gui.dialog.DataFileDialog import DataFileDialog
from silx.io.url import DataUrl
from silx.gui import qt
from est.core.io import read as read_xas, read_frm_file
from est.io import InputType
import logging

_logger = logging.getLogger(__name__)


class _InputType(qt.QWidget):
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(qt.QLabel('input type:'))
        self._inputTypeCB = qt.QComboBox(parent=self)
        for input_type in InputType.values():
            self._inputTypeCB.addItem(input_type)
        self.layout().addWidget(self._inputTypeCB)

        # expose API
        self.currentChanged = self._inputTypeCB.currentIndexChanged

    def getInputType(self):
        """Return the current input type"""
        return InputType.from_value(self._inputTypeCB.currentText())

    def setInputType(self, input_type):
        _input_type = InputType.from_value(input_type)
        idx = self._inputTypeCB.findText(_input_type.value)
        assert idx >= 0
        self._inputTypeCB.setCurrentIndex(idx)


class XASObjectDialog(qt.QWidget):
    """
    Interface used to select inputs for defining a XASObject
    """
    editingFinished = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer, 0, 0)
        self._inputType = _InputType(parent=self)
        self.layout().addWidget(self._inputType, 0, 1)

        self._singleSpectrumDialog = _XASObjectFrmSingleSpectrum(parent=self)
        self.layout().addWidget(self._singleSpectrumDialog, 1, 0, 1, 2)
        self._h5Dialog = _XASObjFrmH5(parent=self)
        self.layout().addWidget(self._h5Dialog, 2, 0, 1, 2)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 3, 0)

        # connect signal / slot
        self._inputType.currentChanged.connect(self._updateWidgetVisibility)
        self._singleSpectrumDialog._inputLe.textChanged.connect(self._editingIsFinished)
        self._h5Dialog.editingFinished.connect(self._editingIsFinished)

        # expose API
        self.getCurrentType = self._inputType.getInputType
        self.setCurrentType = self._inputType.setInputType
        self.setDatFile = self._singleSpectrumDialog.setFileSelected
        self.getDatFile = self._singleSpectrumDialog.getFileSelected
        self.setSpectraUrl = self._h5Dialog.setSpectraUrl
        self.getSpectraUrl = self._h5Dialog.getSpectraUrl
        self.setEnergyUrl = self._h5Dialog.setEnergyUrl
        self.getEnergyUrl = self._h5Dialog.getEnergyUrl
        self.setConfigurationUrl = self._h5Dialog.setConfigurationUrl
        self.getConfigurationUrl = self._h5Dialog.getConfigurationUrl

        # default setting
        self._updateWidgetVisibility()

    def _updateWidgetVisibility(self):
        self._singleSpectrumDialog.setVisible(self._inputType.getInputType() in (InputType.dat_spectrum, InputType.xmu_spectrum))
        self._h5Dialog.setVisible(self._inputType.getInputType() == InputType.hdf5_spectra)

    def buildXASObject(self):
        if self.getCurrentType() in (InputType.dat_spectrum, InputType.xmu_spectrum):
            return read_frm_file(file_path=self._singleSpectrumDialog.getFileSelected())
        elif self.getCurrentType() == InputType.hdf5_spectra:
            spectra_url = self._h5Dialog.getSpectraUrl()
            energy_url = self._h5Dialog.getEnergyUrl()

            def check_url(url_path, name):
                if url_path in (None, ''):
                    raise ValueError(' '.join(('No', name, 'url defined')))
                url = DataUrl(path=url_path)
                if not url.is_valid():
                    raise ValueError(' '.join((name, 'url is invalid. Does the file / path still exists ?')))

            check_url(spectra_url, 'spectra')
            check_url(energy_url, 'energy / channel')
            return read_xas(spectra_url=self._h5Dialog.getSpectraUrl(),
                            channel_url=self._h5Dialog.getEnergyUrl(),
                            config_url=self._h5Dialog.getConfigurationUrl())
        else:
            raise ValueError('unmanaged input type')

    def _editingIsFinished(self, *args, **kwargs):
        self.editingFinished.emit()


class _XASObjectFrmSingleSpectrum(qt.QWidget):
    """
    Interface used to define a XAS object from a single .dat file
    """
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        self.layout().addWidget(qt.QLabel('file', parent=self), 0, 0)
        self._inputLe = qt.QLineEdit('', parent=self)
        self.layout().addWidget(self._inputLe, 0, 1)
        self._selectPB = qt.QPushButton('select', parent=self)
        self.layout().addWidget(self._selectPB, 0, 2)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 2, 0)

        # signal / slot connection
        self._selectPB.pressed.connect(self._selectFile)

        # expose api
        self.getFileSelected = self._inputLe.text

    def _selectFile(self, *args, **kwargs):
        old = self.blockSignals(True)
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFile)
        dialog.setNameFilters(["spectrum (*.dat, *.xmu)", ])

        if not dialog.exec_():
            dialog.close()
            return

        fileSelected = dialog.selectedFiles()
        if len(fileSelected) == 0:
            return
        else:
            assert len(fileSelected) == 1
            self.setFileSelected(fileSelected[0])

        self.blockSignals(old)

    def setFileSelected(self, file_path):
        self._inputLe.setText(file_path)


class _URLSelector(qt.QWidget):
    def __init__(self, parent, name, layout=None, position=None):
        qt.QWidget.__init__(self, parent)
        self.name = name
        if layout is None:
            layout = self.setLayout(qt.QGridLayout())
            position = (0, 0)
        layout.addWidget(qt.QLabel(name + ':', parent=self), position[0], position[1])
        self._qLineEdit = qt.QLineEdit('', parent=self)
        layout.addWidget(self._qLineEdit, position[0], position[1] + 1)
        self._qPushButton = qt.QPushButton('select', parent=self)
        layout.addWidget(self._qPushButton, position[0], position[1] + 2)

        # expose API
        self.getUrlPath = self._qLineEdit.text
        self.setUrlPath = self._qLineEdit.setText

        # connect signal / slot
        self._qPushButton.clicked.connect(self._selectFile)

    def _selectFile(self, *args, **kwargs):
        dialog = DataFileDialog(self)

        if not dialog.exec_():
            dialog.close()
            return None

        if dialog.selectedUrl() is not None:
            self._qLineEdit.setText(dialog.selectedUrl())


class _XASObjFrmH5(qt.QWidget):
    """
    Interface used to define a XAS object from h5 files and data path
    """
    editingFinished = qt.Signal()
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        # spectra url
        self._spectraSelector = _URLSelector(parent=self, name='spectra url',
                                             layout=self.layout(), position=(0, 0))
        # energy / channel url
        self._energySelector = _URLSelector(parent=self, name='energy /channel url',
                                            layout=self.layout(), position=(1, 0))
        # configuration url
        self._configSelector = _URLSelector(parent=self, name='configuration url',
                                            layout=self.layout(),
                                            position=(2, 0))

        # connect signal / slot
        self._spectraSelector._qLineEdit.textChanged.connect(self._editingIsFinished)
        self._energySelector._qLineEdit.textChanged.connect(self._editingIsFinished)
        self._configSelector._qLineEdit.textChanged.connect(self._editingIsFinished)

    def getSpectraUrl(self):
        return self._spectraSelector.getUrlPath()

    def getEnergyUrl(self):
        return self._energySelector.getUrlPath()

    def getConfigurationUrl(self):
        return self._configSelector.getUrlPath()

    def setSpectraUrl(self, url):
        self._spectraSelector.setUrlPath(url)

    def setEnergyUrl(self, url):
        self._energySelector.setUrlPath(url)

    def setConfigurationUrl(self, url):
        self._configSelector.setUrlPath(url)

    def _editingIsFinished(self, *args, **kwargs):
        self.editingFinished.emit()


if __name__ == '__main__':
    app = qt.QApplication([])
    dialog = XASObjectDialog()
    dialog.show()
    app.exec_()
